/*
License for ScriptEd

Copyright 2005 by Steven Berry <berrys66@yahoo.co.uk>

All Rights Reserved 

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that 
both that copyright notice and this permission notice appear in 
supporting documentation. 

STEVEN BERRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS, IN NO EVENT SHALL STEVEN BERRY BE LIABLE FOR ANY 
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, 
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
OR PERFORMANCE OF THIS SOFTWARE. 
*/

/*

Author:         Steven Berry.
Last modified:  10/12/2001
Version:        1.0

Implementation file for XML C++ class.

All effort has been made to make this class function according to 
the W3C XML specification - 

"Extensible Markup Language (XML) 1.0 (Second Edition) W3C Recommendation 6 October 2000"

This document may be found at:

http://www.w3.org/TR/2000/REC-xml-20001006


There are a number of features of the XML specification that either are not implemented
or do not work as advertised:

  XML Parser - Incorrectly parses "Mix content" type elements

  DTD Parser - Incorrectly interprets mixed content declarations of the form (#PCDATA,element)+
             - Not implemented: NOTATION attribute types.
             - Not implemented: Tokenised attribute declarations.
             - Limitation: Conditional sections cannot be nested.

  Validator  - Probably numerous bugs.

  Doc producer - Probably numerous bugs.

*/

#include <windows.h>
#include <stdio.h>

#include "xmlparse.h"
#include "xmlerror.h"


XML::XML( void ){ initialise() ; }


XML::XML( long dSize, char *Data, int *Status )
{

  initialise() ;
  if ( (*Status = LoadData( dSize, Data )) == 0 ) Parse() ;

}


XML::~XML( void )
{

  XMLWipe() ;
  DTDWipe() ;
  DocWipe() ;

}


void XML::XMLWipe( void )
{

  long i ;


  if ( XMLData != NULL ) delete XMLData ;
  if ( Top.NextIndex != NULL )
  {
    CurrentNode = &Top ;
    for ( i = 0 ; i < Index ; i++ )
      CurrentNode = CurrentNode->NextIndex ;
    for ( ; i > 0 ; i-- )
    {
      CurrentNode = CurrentNode->PrevIndex ;
      if ( CurrentNode->AttrValues != NULL ) delete CurrentNode->AttrValues ;
      delete CurrentNode->NextIndex ;
    }
  }

  ZapXMLNode( &Top ) ;
  SystemState &= ~SYS_STATE_XMLLOADED ;
  SystemState &= ~SYS_STATE_XMLPARSED ;

}


void XML::DTDWipe( void )
{

  long i ;


  if ( DTD.NextIndex != NULL )
  {
    CurrentDTDNode = &DTD ;
    for ( i = 0 ; i < DTDIndex ; i++ )
      CurrentDTDNode = CurrentDTDNode->NextIndex ;
    for ( ; i > 0 ; i-- )
    {
      CurrentDTDNode = CurrentDTDNode->PrevIndex ;
      if ( CurrentDTDNode->NextIndex->Type = NODE_ATTLIST && CurrentDTDNode->NextIndex->Attributes != NULL && CurrentDTDNode->NextIndex->FreeAtt == 1 )
        free(CurrentDTDNode->NextIndex->Attributes) ;
      delete CurrentDTDNode->NextIndex ;
    }
  }
  if ( DTDData != NULL ) delete DTDData ;

}


void XML::DocWipe( void )
{

  long i ;

  if ( Doc.NextIndex != NULL )
  {
    CurrentDocNode = &Doc ;
    for ( i = 0 ; i < DocIndex ; i++ )
      CurrentDocNode = CurrentDocNode->NextIndex ;
    for ( ; i > 0 ; i-- )
    {
      CurrentDocNode = CurrentDocNode->PrevIndex ;
      if ( CurrentDocNode->NextIndex->Value != NULL ) free( CurrentDocNode->NextIndex->Value ) ;
      if ( CurrentDocNode->NextIndex->Type == NODE_ATTRVAL )
      {
        if ( CurrentDocNode->NextIndex->Name != NULL ) free(CurrentDocNode->NextIndex->Name) ;
      }
      else
        if ( DocGotDTD == 0 && CurrentDocNode->NextIndex->Name != NULL ) free(CurrentDocNode->NextIndex->Name) ;
      delete CurrentDocNode->NextIndex ;
    }
  }
  if ( Doc_DocType != NULL ) free( Doc_DocType ) ;
  if ( Doc_XMLVersion != NULL ) free( Doc_XMLVersion ) ;
  Doc_DocType = NULL ;
  Doc_XMLVersion = NULL ;
  DocIndex = 0 ;
  if ( DocGotDTD == 0 )
  {
    if ( Doc.Name != NULL ) free(Doc.Name) ;
  }
  DocGotDTD = 0 ;
  ZapDocNode( &Doc ) ;
  SystemState &= ~SYS_STATE_DOCINIT ;

}


void XML::initialise( void )
{

  ZapXMLNode( &Top ) ;
  Index = 0 ;
  Top.Type = NODE_UNDEFINED ;
  CurrentNode = &Top ;
  LastNode = CurrentNode ;
  XMLData = NULL ;
  XMLVersion = NULL ;
  DocType = NULL ;
  DTDData = NULL ;
  DTD.Index = 0 ;
  DTD.Name = NULL ;
  DTD.NextNode = NULL ;
  DTD.PrevNode = NULL ;
  DTD.Type = DTD_TOP ;
  DTD.Conditional = 0 ;
  DTD.NodeData = NULL ;
  DTD.ContentModel = 0 ;
  DTD.ContentType = 0 ;
  DTD.ContModel = NULL ;
  DTD.ContType = NULL ;
  DTD.NextIndex = NULL ;
  DTD.Occurence = 0 ;
  DTD.Parent = NULL ;
  DTD.PrevIndex = NULL ;
  DTD.FirstChild = NULL ;
  DTD.AttrReq = 0 ;
  DTD.AttrTok = 0 ;
  DTD.AttrType = 0 ;
  DTD.FirstAttr = NULL ;
  DTD.Attributes = NULL ;
  CurrentDTDNode = &DTD ;
  DTDVersion = NULL ;
  DTDIndex = 0 ;
  CondLevel = 0 ;
  SystemState = 0 ;
  DocIndex = 0 ;
  CurrentDocNode = &Doc ;
  ZapDocNode( &Doc ) ;
  LastDocNode = NULL ;
  LastDocIndex = &Doc ;
  XMLDoc = NULL ;
  Doc_DocType = NULL ;
  Doc_XMLVersion = NULL ;
  DocGotDTD = 0 ;
  InitEscapes() ;

}


void XML::InitEscapes( void )
{

  esc[0].eSeq = "&amp;" ;
  esc[0].eChar = '&' ;
  esc[0].eLen = 5 ;

  esc[1].eSeq = "&gt;" ;
  esc[1].eChar = '>' ;
  esc[1].eLen = 4 ;

  esc[2].eSeq = "&lt;" ;
  esc[2].eChar = '<' ;
  esc[2].eLen = 4 ;

  esc[3].eSeq = "&quot;" ;
  esc[3].eChar = '"' ;
  esc[3].eLen = 6 ;

  esc[4].eSeq = "&apos;" ;
  esc[4].eChar = '\'' ;
  esc[4].eLen = 6 ;

  esc[5].eSeq = "&shy;" ;
  esc[5].eChar = (char)0xAD ;
  esc[5].eLen = 5 ;

}


int XML::LoadData( long dSize, char *Data )
{

  long i ;
  char *p, *q ;


  if ( dSize == 0 || Data == NULL )
    return( E_XML_NODATA ) ;

  XMLData = new char [dSize + 1] ;
  for ( i = 0, p = XMLData, q = Data ; i < dSize ; i++ )
    *p++ = *q++ ;

  while( *p != '>' ) p-- ;
  p++ ;
  *p = 0 ;

  
  XMLDataLength = p - XMLData ;

  SystemState |= SYS_STATE_XMLLOADED ;
  return( 0 ) ;

}


int XML::LoadDTD( long dSize, char *Data )
{

  long i ;
  char *p, *q ;


  if ( dSize == 0 || Data == NULL )
    return( E_XML_NODATA ) ;

  DTDData = new char [dSize + 1] ;
  for ( i = 0, p = DTDData, q = Data ; i < dSize ; i++ )
    *p++ = *q++ ;

  while( *p != '>' ) p-- ;
  p++ ;
  *p = 0 ;

  
  DTDDataLength = p - DTDData ;

  SystemState |= SYS_STATE_DTDLOADED ;

  return( 0 ) ;

}


int XML::ParseDTD( void ){ return( ParseDTD1( DTDData, DTDDataLength ) ) ; }


int XML::ParseDTD1( char *DataStart, long DataLength )
{

  char *p ;
  long ecrv ;
  int ret ;


  if ( SYS_DTDREADY(SystemState) ) return(E_DTD_ALREADY) ;
  if ( ! SYS_DTDLOADED ) return(E_DTD_NODTD) ;

  p = SkipWhiteSpace( DataStart ) ;

  if ( *p == '<' && *(p+1) == '?' )
  {
    p = SkipWhiteSpace( p+2 ) ;
    if ( (*p & ~32)      == 'X' &&
         (*(p+1) & ~32 ) == 'M' &&
         (*(p+2) & ~32 ) == 'L' &&
          *(p+3)         == 32
       )
    {
      p = SkipWhiteSpace( p+4 ) ;
      DTDVersion = p ;
      while ( *p && *p != '?' ) p++ ;
      if ( *p != '?' ) DTDVersion = p ;
      *p++ = 0 ;
    }
    else
    {
      *p = 0 ;
      XMLVersion = p++ ;
    }
    while( *p && *p != '>' ) p++ ;
    p = SkipWhiteSpace( p+1 ) ;
  }

  for ( ; ; )
  {
    if ( *p == '<' && *(p+1) == '!' )
    {
      p = SkipWhiteSpace( p+2 ) ;
      if ( strnicmp( (const char *)p, "ENTITY ", 7 ) == 0 )
      {
        p += 7 ;
        p = SkipWhiteSpace(p) ;
        p += AddDTDNode( p, NODE_ENTITY ) ;
      }
      else
      {
        if ( strnicmp( (const char *)p, "ELEMENT ", 8 ) == 0 )
        {
          p += 8 ;
          p = SkipWhiteSpace(p) ;
          p += AddDTDNode( p, NODE_ELEMENT ) ;
        }
        else
        {
          if ( strnicmp( (const char *)p, "ATTLIST ", 8 ) == 0 )
          {
            p += 8 ;
            p = SkipWhiteSpace(p) ;
            p += AddDTDNode( p, NODE_ATTLIST ) ;
          }
          else
          {
            if ( strnicmp( (const char *)p, S_COMMENT, S_LEN_COMMENT ) == 0 )
            {
              p += S_LEN_COMMENT ;
              while ( strnicmp( E_COMMENT, (const char *)p, E_LEN_COMMENT ) != 0 )
              {
                p++ ;
                if ( p - DataStart > DataLength )
                  return(E_DTD_INVALID) ;
              }
              p += E_LEN_COMMENT ;
            }
            else
            {
              if ( strnicmp( (const char *)p, S_COND_SECT, S_LEN_COND_SECT ) == 0 )
              {
                p += S_LEN_COND_SECT ;
                ecrv = EvalCondSect( p ) ;
                if ( ecrv != -1 )
                  p += ecrv ;
                else
                  return(E_DTD_INVALID) ;
              }
              else
                return(E_DTD_INVALID_CHAR) ;
            }
          }
        }
      }
    }
    else
      return(E_DTD_INVALID_CHAR) ;
    p = SkipWhiteSpace(p) ;
    if ( p - DataStart >= DataLength )
      break ;
  }

  if ( (ret = ResolveDTD()) == 0 )
    SystemState |= SYS_STATE_DTDPARSED ;

  return(ret) ;
}


long XML::EvalCondSect( char *eptr )
{

  char *p, *q, *r ;
  struct _BaseDTD *nodeptr ;
  int replace ;


  p = eptr ;
  p = SkipWhiteSpace(p) ;
  if ( *p == '%' )
  {
    replace = 1 ;
    q = SkipWhiteSpace(p + 1) ;
    p = q + 1 ;
    while ( *p && *p != ';' ) p++ ;
    if ( *p != ';' ) return(-1) ;
    *p++ = 0 ;
    p = SkipWhiteSpace(p) ;
    if ( *p != '[' ) return(-1) ;
    p++ ;
  }
  else
  {
    replace = 0 ;
    q = p ;
    while( *p && *p != '[' ) p++ ;
    if ( *p != '[' ) return(-1) ;
    *p = 0 ;
    p++ ;
  }

  r = p ;

  while( strnicmp( (const char *)r, E_COND_SECT, E_LEN_COND_SECT ) != 0 )
  {
    if ( *r == 0 ) return(-1) ;
    r++ ;
  } 

  if ( replace )
  {
    for( nodeptr = &DTD ; nodeptr->NextNode != NULL ; )
    {
      nodeptr = nodeptr->NextNode ;
      if ( nodeptr->Type == NODE_ENTITY )
      {
        if ( nodeptr->Conditional == 1 )
        {
          if ( strcmp( (const char *)nodeptr->Name, q ) == 0 )
          {
            q = nodeptr->NodeData ;
            break ;
          }
        }      
      }
    }
  }

  if ( stricmp( (const char *)q, COND_INCLUDE ) == 0 ||
       stricmp( (const char *)q, COND_IGNORE ) == 0 )
  {
    if ( stricmp( (const char *)q, COND_INCLUDE ) == 0 )
    {
      replace = ParseDTD1( p, r - p ) ;
      if ( replace != 0 ) return(-1) ;
    }
  }
  else
    return(-1) ;

  return( r - eptr + E_LEN_COND_SECT ) ;

}


long XML::AddDTDNode( char *NodeData, int NodeType )
{

  char *p, *q, c ;
  struct _BaseDTD *nodeptr ;


  nodeptr = new struct _BaseDTD ;
  DTDIndex++ ;
  CurrentDTDNode->NextNode = nodeptr ;
  nodeptr->Conditional = 0 ;
  nodeptr->Index = DTDIndex ;
  nodeptr->PrevNode = CurrentDTDNode ;
  nodeptr->Type = NodeType ;
  nodeptr->NextNode = NULL ;
  nodeptr->Name = NULL ;
  nodeptr->NodeData = NULL ;
  nodeptr->NextIndex = NULL ;
  nodeptr->PrevIndex = CurrentDTDNode ;
  CurrentDTDNode->NextIndex = nodeptr ;
  nodeptr->ContentModel = 0 ;
  nodeptr->ContentType = 0 ;
  nodeptr->ContModel = NULL ;
  nodeptr->ContType = NULL ;
  nodeptr->Occurence = 0 ;
  nodeptr->Parent = &DTD ;
  nodeptr->FirstChild = NULL ;
  nodeptr->AttrReq = 0 ;
  nodeptr->AttrTok = 0 ;
  nodeptr->AttrType = 0 ;
  nodeptr->FirstAttr = NULL ;
//  nodeptr->FirstEnum = NULL ;
  nodeptr->Attributes = NULL ;

  p = NodeData ;

  switch( NodeType )
  {
    case NODE_ENTITY:
      if ( *p == '%' )
      {
        nodeptr->Conditional = 1 ;
        p++ ;
        p = SkipWhiteSpace(p) ;
      }
    case NODE_ATTLIST:
    case NODE_ELEMENT:
      nodeptr->Name = p ;
      p = FindWhiteSpace(p) ;
      *p++ = 0 ;
      p = SkipWhiteSpace(p) ;
      nodeptr->NodeData = p ;
      p = FindClosing(p) ;
      *p++ = 0 ;
      break ;
  }

//  if ( NodeType == NODE_ENTITY && nodeptr->Conditional == 1 )
  if ( NodeType == NODE_ENTITY )
  {
    q = nodeptr->NodeData ;
    if ( *q == 34 || *q == '\'') 
    {
      c = *q ;
      nodeptr->NodeData = q + 1 ;
      q++ ;
//      while ( *q != 34 && *q != '\'' ) q++ ;
      while ( *q != c ) q++ ;
      *q = 0 ;
    }
  }
  CurrentDTDNode = nodeptr ;
  LastDTDNode = nodeptr ;
  LastDTDIndex = nodeptr ;
  return(p - NodeData) ;

}


void XML::ZapNode( struct _BaseDTD *nodeptr )
{

  char *p ;
  int i ;

  for ( i = 0, p = (char *)nodeptr ; i < sizeof(struct _BaseDTD) ; i++ )
    *p++ = 0 ;
  DTDIndex++ ;
  nodeptr->Index = DTDIndex ;
  nodeptr->PrevIndex = LastDTDIndex ;
  LastDTDIndex->NextIndex = nodeptr ;
  LastDTDIndex = nodeptr ;

}


void XML::ZapDocNode( struct _Node *nodeptr )
{

  char *p ;
  int i ;

  for ( i = 0, p = (char *)nodeptr ; i < sizeof(struct _Node) ; i++ )
    *p++ = 0 ;

  if ( nodeptr != &Doc )
  {
    DocIndex++ ;
    nodeptr->Index = DocIndex ;
    nodeptr->PrevIndex = LastDocIndex ;
    LastDocIndex->NextIndex = nodeptr ;
    LastDocIndex = nodeptr ;
  }

}


void XML::ZapXMLNode( struct _Node *nodeptr )
{

  char *p ;
  int i ;

  for ( i = 0, p = (char *)nodeptr ; i < sizeof(struct _Node) ; i++ )
    *p++ = 0 ;

  if ( nodeptr != &Top )
  {
    Index++ ;
    nodeptr->Index = Index ;
    nodeptr->PrevIndex = LastNode ;
    LastNode->NextIndex = nodeptr ;
    LastNode = nodeptr ;
  }

}


int XML::ResolveDTD( void )
{

  int r ;


  for( CurrentDTDNode = &DTD ; CurrentDTDNode->NextNode != NULL ; )
  {
    CurrentDTDNode = CurrentDTDNode->NextNode ;
    switch( CurrentDTDNode->Type )
    {
      case NODE_ENTITY:
        r = 0 ;
        break ;

      case NODE_ELEMENT:
        r = ResolveElement( CurrentDTDNode ) ;
        break ;
      case NODE_ATTLIST:
        r = ResolveAttlist( CurrentDTDNode ) ;
        break ;
    }
    if ( r != 0 ) return(r) ;
  }
  return(0) ;

}

/*
char *XML::ResolveEntity( const char *EntityName )
{

  struct _BaseDTD *nodeptr ;
  char *p, *q ;


  for ( nodeptr = &DTD ; nodeptr->NextNode != NULL ; )
  {
    nodeptr = nodeptr->NextNode ;
    if ( nodeptr->Type == NODE_ENTITY )
    {
      if ( strcmp( EntityName, (const char *)nodeptr->Name ) == 0 )
      {
        p = nodeptr->NodeData ;
        if ( *p == '%' )
        {
          q = strdup( (p+1) ) ;
          q[strlen(p) - 2] = 0 ;
          p = ResolveEntity( (const char *)q ) ;
          free(q) ;
          return(p) ;
        }
        else
          return(p) ;
      }
    }
  }
  return(NULL) ;

}
*/

char *XML::ResolveEntity( const char *EntityName )
{

  struct _BaseDTD *nodeptr ;
  char *p, *q, *r, *s ;


  for ( nodeptr = &DTD ; nodeptr->NextNode != NULL ; )
  {
    nodeptr = nodeptr->NextNode ;
    if ( nodeptr->Type == NODE_ENTITY )
    {
      if ( strcmp( EntityName, (const char *)nodeptr->Name ) == 0 )
      {
        p = nodeptr->NodeData ;
        r = strdup(p) ;
        while ( ( p = strstr( (const char *)r, "%" )) != NULL )
        {
          q = strdup( p ) ;
          s = q + 1 ;
          p = s ;
          while ( *p && *p != ' ' && *p != ';' ) p++ ;
          if ( *p != ';' )
          {
            free(q) ;
            free(r) ;
            return(NULL) ;
          }
          *++p = 0 ;
          p = strdup(s) ;
          p[strlen(p) - 1] = 0 ;
          s = ResolveEntity( (const char *)p ) ;
          if ( s != NULL )
          {
            free(p) ;
            p = Replace( r, q, s ) ;
            free(q) ;
            free(r) ;
            r = p ;
          }
          else
          {
            free(q) ;
            free(p) ;
            free(r) ;
            return(NULL) ;
          }
        }
        return(r) ;
      }
    }
  }
  return(NULL) ;

}


int XML::ResolveElement( struct _BaseDTD *ElementPtr )
{

  int ret ;


  ElementPtr->ContentType = GetElementContentType( (const char *)ElementPtr->NodeData ) ;
  switch ( ElementPtr->ContentType )
  {
    case ELEMENT_CONTENT_UNKNOWN:
      ret = E_DTD_CONTENT_TYPE ;
      break ;

    case ELEMENT_CONTENT_MIXED:
      ret = ResolveElementMixed( ElementPtr ) ;
      break ;

    case ELEMENT_CONTENT_KIDS:
      ret = ResolveElementKids( ElementPtr ) ;
      break ;

    default:
      ret = 0 ;
      break ;
  }
  
  return(ret) ;

}


int XML::ResolveElementMixed( struct _BaseDTD *ElementPtr )
{

  int rec ;
  char *resed, *mem ;


  if ( (rec = ResolveElementCommon( ElementPtr, &resed, &mem )) != 0 )
    return(rec) ;
  rec = ResolveElementDetail( ElementPtr, resed + 1, strlen((const char *)resed) - 2 ) ;
  free(mem) ;
  return(0) ;

}


int XML::ResolveElementKids( struct _BaseDTD *ElementPtr )
{

  int rec ;
  char *resed, *mem ;


  if ( (rec = ResolveElementCommon( ElementPtr, &resed, &mem )) != 0 )
    return(rec) ;
  rec = ResolveElementDetail( ElementPtr, resed + 1, strlen((const char *)resed) - 2 ) ;
  free(mem) ;
  return(rec) ;

}


int XML::ResolveElementDetail( struct _BaseDTD *ParentPtr, char *start, long reslen )
{

  char *p, *q, *r, *s, *t ;
  struct _BaseDTD *nodeptr, *elnode, *contnode, *contlast ;
  int nestlevel = 0 ;
  int ret = 0, out = 0 ;


  for ( p = start, contnode = NULL ; p - start < reslen ; )
  {
    nodeptr = new struct _BaseDTD ;
    ZapNode( nodeptr ) ;
    nodeptr->Parent = ParentPtr ;
//    LastDTDNode->NextIndex = nodeptr ;
//    nodeptr->PrevIndex = LastDTDNode ;
    if ( ParentPtr->FirstChild == NULL ) ParentPtr->FirstChild = nodeptr ;
    if ( *p == '(' )
    {
      out = 0 ;
      contnode = nodeptr ;
      contlast = LastDTDNode ;
      nodeptr->Type = NODE_CONTENT_MODEL ;
      if ( LastDTDNode->Parent == nodeptr->Parent )
      {
        if ( nodeptr->Parent->FirstChild != NULL )
        {
          nodeptr->ContentModel = LastDTDNode->ContentModel ;
          nodeptr->PrevNode = LastDTDNode ;
          LastDTDNode->NextNode = nodeptr ;
        }
      }
      p++ ;
      for ( q = p ; q - start < reslen ; q++ )
      {
        if ( out == 1 ) break ;
        if ( *q == '(' ) nestlevel++ ;
        if ( *q == ')' )
        {
          if ( nestlevel == 0 )
          {
            LastDTDNode = nodeptr ;
            if ( (ret = ResolveElementDetail( nodeptr, p, q - p )) != 0 )
              return(ret) ;
            out = 1 ;
            p = q + 1 ;
            if ( *p == '*' || *p == '?' || *p == '+' )
            {
              switch (*p)
              {
                case '*':
                  nodeptr->Occurence = ELEMENT_ZERO_OR_MORE ;
                  break ;
                case '+':
                  nodeptr->Occurence = ELEMENT_ONE_OR_MORE ;
                  break ;
                case '?':
                  nodeptr->Occurence = ELEMENT_ZERO_OR_ONE ;
                  break ;
              }
              p++ ;
            }
            else
              nodeptr->Occurence = ELEMENT_ONLY_ONCE ;
            if ( contlast->Parent == nodeptr->Parent )
            {
              if ( nodeptr->Parent->FirstChild == NULL )
              {
                nodeptr->ContentModel = DTDNodeContentType(*p) ;
                if ( nodeptr->ContentModel >= 1000 )
                  return(E_DTD_CONTENT_BAD) ;
              }
            }
            else
            {
              nodeptr->ContentModel = DTDNodeContentType(*p) ;
              if ( nodeptr->ContentModel >= 1000 )
                return(E_DTD_CONTENT_BAD) ;
              LastDTDNode = nodeptr ;
            }   
            if ( *p == ',' || *p == '|' ) p++ ;
            contnode = nodeptr ;
          }
          else
            nestlevel-- ;
        }
      }
      if ( q - start > reslen )
        return(E_DTD_ELEMENT_GROUP) ;
    }
    else
    {
      nodeptr->Type = NODE_CONTENT_TYPE ;
      for ( q = p ; ; q++)
      {
        if ( *q == ',' || *q == '|' || *q == ')' ) break ;
        if ( q - start >= reslen )
          return(E_DTD_ELEMENT_GROUP) ;
      }
      r = (char *)malloc( q - p + 1 ) ;
      s = r ;
      t = p ;
      while ( t < q ) *s++ = *t++ ;
      *s = 0 ;
      if ( *q == ',' ) 
        nodeptr->ContentModel = CONTENT_MODEL_SEQ ;
      if ( *q == '|' ) 
        nodeptr->ContentModel = CONTENT_MODEL_CHOICE ;
      switch( *(q-1) )
      {
        case '*':
          nodeptr->Occurence = ELEMENT_ZERO_OR_MORE ;
          *(s-1) = 0 ;
          break ;
        case '+':
          nodeptr->Occurence = ELEMENT_ONE_OR_MORE ;
          *(s-1) = 0 ;
          break ;
        case '?':
          nodeptr->Occurence = ELEMENT_ZERO_OR_ONE ;
          *(s-1) = 0 ;
          break ;
        default:
          nodeptr->Occurence = ELEMENT_ONLY_ONCE ;
          break ;
      }
      if ( strcmp( (const char *)r, "#PCDATA" ) == 0 )
      {
        elnode = CONTENT_IS_PCDATA ;
        free(r) ;
      }
      else
      {
        elnode = FindElementPtr( r ) ;
        free(r) ;
        if ( elnode == NULL )
          return(E_DTD_ELEMENT_BAD) ;
      }
      nodeptr->ContType = elnode ;
      if ( contnode == NULL )
      {
        if ( nodeptr->Parent->FirstChild != nodeptr )
        {
          nodeptr->PrevNode = LastDTDNode ;
          LastDTDNode->NextNode = nodeptr ;
        }
      }
      else
      {
        nodeptr->PrevNode = contnode ;
        contnode->NextNode = nodeptr ;
        contnode = NULL ;
      }
      if ( *q == ')' )
      {
        if ( nodeptr->PrevNode != NULL )
          nodeptr->ContentModel = nodeptr->PrevNode->ContentModel ;
        else
          nodeptr->ContentModel = CONTENT_MODEL_SINGLE ;
      }
      LastDTDNode = nodeptr ;
      p = q + 1 ;
    }
  }  

  return(ret) ;

}


int XML::DTDNodeContentType( char c )
{

  switch (c)
  {
    case ',':
      return(CONTENT_MODEL_SEQ) ;
      break ;
    case '|':
      return(CONTENT_MODEL_CHOICE) ;
      break ;
    case ')':
      return(CONTENT_MODEL_SINGLE) ;
      break ;
    }
    return(E_DTD_CONTENT_BAD) ;
}


int XML::GetContentType( void ) { return( CurrentDTDNode->ContentType ) ; }


struct XML::_BaseDTD *XML::FindElementPtr( char *elname )
{

  struct _BaseDTD *nodeptr ;


  for ( nodeptr = &DTD ; nodeptr->NextNode != NULL ; )
  {
    nodeptr = nodeptr->NextNode ;
    if ( nodeptr->Type == NODE_ELEMENT )
    {
      if ( strcmp( (const char *)elname, (const char *)nodeptr->Name ) == 0 )
        return( nodeptr ) ;
    }
  }
  return( (struct _BaseDTD *)NULL ) ;

}


int XML::ResolveElementCommon( struct _BaseDTD *ElementPtr, char **res, char **ptr )
{

  char *p, *q, *r, *s, *t, *u, *v ;
  char *epd ;
  int free_it = 0 ;


  epd = v = ElementPtr->NodeData ;
  while ( (p = strstr( (const char *)epd, "%" )) != NULL )
  {
    q = strstr((const char *)epd, ";" ) ;
    if ( q == NULL )
      return(E_DTD_ENTITY_SUBS) ;
    else
    {
      r = (char *)malloc( q - p + 2 ) ;
      t = r ;
      while ( p <= q ) *t++ = *p++ ;
      *t = 0 ; 
      u = strdup( (r+1) ) ;
      u[strlen(r) - 2] = 0 ;
      s = ResolveEntity( (const char *)u ) ;
      if ( s == NULL )
      {
        if ( free_it == 1 ) free(epd) ;
        free(u) ;
        free(r) ;
        return(E_DTD_ENTITY_BAD) ;
      }
      v = Replace( epd, r, s ) ;
      if ( free_it == 1 ) free(epd) ;
      epd = v ;
      free_it = 1 ;
      free(u) ;
      free(r) ;
    }
  }

  r = (char *)malloc( strlen(epd) + 1 ) ;
  p = epd ;
  q = r ;
  while ( *p )
  {
    if ( *p > 32 && *p < 127 )
      *q++ = *p ;
    p++ ;
  }
  *q = 0 ;
  if ( free_it == 1 ) 
    free(epd) ;
  epd = r ;
  if ( r[0] == '(' && r[1] == '(' &&
       r[strlen((const char *)r)-2] == ')' && r[strlen((const char *)r)-1] == ')' )
  {
    epd = r + 1 ;
    r[strlen((const char *)r)-1] = 0 ;
  }

  *res = epd ;
  *ptr = r ;
  return(0) ;

}


int XML::GetElementContentType( const char *ElementData )
{

  const char *p ;
  char *q, *r ;
  int ctype ;


  r = (char *)malloc( strlen(ElementData) + 1 ) ;
  p = ElementData ;
  q = r ;
  while ( *p )
  {
    if ( *p > 32 && *p < 127 )
      *q++ = *p ;
    p++ ;
  }
  *q = 0 ;

  if ( strcmp( (const char *)r, "(#PCDATA)" ) == 0 )
    ctype = ELEMENT_CONTENT_PCDATA ;
  else
    if ( strcmp( (const char *)r, "ANY" ) == 0 )
      ctype = ELEMENT_CONTENT_ANY ;
    else
      if ( strcmp( (const char *)r, "EMPTY" ) == 0 )
        ctype = ELEMENT_CONTENT_EMPTY ;
      else
        if ( strstr( (const char *)r, "#PCDATA" ) != NULL )
          ctype = ELEMENT_CONTENT_MIXED ;
        else
          if ( strstr( (const char *)r, "(" ) != NULL)
            ctype = ELEMENT_CONTENT_KIDS ;
          else
            ctype = ELEMENT_CONTENT_UNKNOWN ;

  free(r) ;
  return(ctype) ;

}


int XML::ResolveAttlist( struct _BaseDTD *AttlistPtr )
{

  char *p, *q, *s ;
  struct _BaseDTD *Element, *nodeptr, *tmpnode, *tmpnode2 ;
  int ret ;
  int gotquote, out ;

  
  if (AttlistPtr->Index == 157 )
    ret = 0 ;

  Element = FindElementPtr( AttlistPtr->Name ) ;
  if ( Element == NULL )   
    return(0) ;           /* Ignore declarations for non-existent elements. This is not an error according to the XML spec! */

  if ( Element->Attributes != NULL ) return(0) ; /* Ignore duplicate declarations. This, apparently isn't an error either! */

  ret = ResolveAttlistEntities( AttlistPtr ) ;
  if ( ret != 0 ) return(ret) ;

  for( p = AttlistPtr->Attributes ; ; )
  {
    nodeptr = new struct _BaseDTD ;
    ZapNode( nodeptr ) ;
    nodeptr->Parent = Element ;
    if ( Element->FirstAttr == NULL )
      Element->FirstAttr = nodeptr ;
    else
    {
      if ( tmpnode == NULL )
      {
        nodeptr->PrevNode = LastDTDNode ;
        LastDTDNode->NextNode = nodeptr ;
      }
      else
      {
        nodeptr->PrevNode = tmpnode ;
        tmpnode->NextNode = nodeptr ;
      }
    }
//    nodeptr->PrevIndex = LastDTDNode ;
//    LastDTDNode->NextIndex = nodeptr ;
    LastDTDNode = nodeptr ;
    tmpnode = NULL ;
    nodeptr->Type = NODE_ATTRIBUTE_DEF ;
    q = p ;
    while ( *p && *p != 32 ) p++ ;
    if ( *p == 0 ) return(E_DTD_ATTR_NAME) ;
    *p++ = 0 ;
    nodeptr->Name = q ;
    nodeptr->AttrType = DTDNodeAttributeType( p, &nodeptr->AttrTok ) ;
    if ( nodeptr->AttrType == E_DTD_ATTR_TYPE )
      return(E_DTD_ATTR_TYPE) ;
    if ( nodeptr->AttrType == ATTRIBUTE_TYPE_ENUM )
    {
      nodeptr->AttrReq = ATTR_ENUM ;
      tmpnode = nodeptr ;
      if ( *p != '(' ) return(E_DTD_ATTR_ENUM) ;
      p = SkipWhiteSpace(p+1) ;
      out = 0 ;
      gotquote = 0 ;
      while ( out == 0 && *p != ')' )
      {
        if ( tmpnode->FirstChild == NULL || *p == '|' || gotquote == '|' )
        {
          nodeptr = new struct _BaseDTD ;
          ZapNode( nodeptr ) ;
          if ( *p == '|' ) p = SkipWhiteSpace(p+1) ;
        }
        gotquote = 32 ;
        if ( *p == '"' || *p == '\'' || *p == '|' )
        {
          gotquote = *p ;
          p++ ;
        }
        q = p ;
        while ( *p && *p != ')' && ( ( gotquote != 32 && *p != gotquote ) || *p != '|' ) ) p++ ;
        switch( *p )
        {
          case 0:
            return(E_DTD_ATTR_ENUM) ;
            break ;

          case ')':
            out = 1 ;
            break ;

          case '|':
            if ( gotquote == '"' || gotquote == '\'' )
              return(E_DTD_ATTR_QUOTE) ;
            s = p - 1 ;
            while( *s == 32 ) s-- ;
            s++ ;
            *s = 0 ;
            gotquote = '|' ;
            break ;
        }
        *p++ = 0 ;
 //       nodeptr->PrevIndex = LastDTDNode ;
 //       LastDTDNode->NextIndex = nodeptr ;
        if ( tmpnode->FirstChild == NULL )
          tmpnode->FirstChild = nodeptr ;
        else
        {
          LastDTDNode->NextNode = nodeptr ;
          nodeptr->PrevNode = LastDTDNode ;    
        }
        nodeptr->Parent = tmpnode ;
        nodeptr->Type = NODE_ATTRIBUTE_VAL ;
        LastDTDNode = nodeptr ;
        nodeptr->NodeData = q ;
        p = SkipWhiteSpace(p) ;
      }
      if ( *p == ')' ) p = SkipWhiteSpace(p+1) ;
      if ( *p == 0 ) return(E_DTD_ATTR_ENUM) ;
      tmpnode->AttrReq = DTDNodeAttributeReq( p ) ;
//      if ( tmpnode->AttrReq == E_DTD_ATTR_REQ || tmpnode->AttrReq == ATTR_FIXED )
      if ( tmpnode->AttrReq == E_DTD_ATTR_REQ ) // || tmpnode->AttrReq == ATTR_FIXED )
        return(E_DTD_ATTR_ENUM_DEF) ;
      if ( tmpnode->AttrReq == ATTR_FIXED )
      {
        p += LEN_ATTR_FIXED_STR ;
        if ( *p != 0 ) p++ ;
      }
      if ( tmpnode->AttrReq == ATTR_IMPLIED ) 
      {
        p += LEN_ATTR_IMPLIED_STR ;
        if ( *p != 0 ) p++ ;
      }
      else
      {
        gotquote = 32 ;
        if ( *p == '"' || *p == '\'')
        {
          gotquote = *p ;
          p++ ;
        }
        q = p ;
        while ( *p && *p != gotquote ) p++ ;
        *p++ = 0 ;
        if ( tmpnode->AttrReq == ATTR_VALUE || tmpnode->AttrReq == ATTR_FIXED )
        {
          tmpnode->AttrReq = 0 ;
          for ( tmpnode2 = tmpnode->FirstChild, out = 0 ; ; )
          {
            if ( strcmp( (const char *)tmpnode2->NodeData, (const char *)q ) == 0 )
            {
              out = 1 ;
              break ;
            }
            if ( tmpnode2->NextNode != NULL )
              tmpnode2 = tmpnode2->NextNode ;
            else
              break ;
          }
          if ( out == 1 )
            tmpnode->NodeData = q ;
          else
            return(E_DTD_ATTR_ENUM_DEF) ;
        }
        p = SkipWhiteSpace(p) ;
      }
    }
    else
    {
      while ( *p && *p != 32 ) p++ ;
      if ( *p == 0 ) return(E_DTD_ATTR_REQ) ;
      p++ ;
      nodeptr->AttrReq = DTDNodeAttributeReq( p ) ;
      if ( nodeptr->AttrReq == E_DTD_ATTR_REQ )
        return(E_DTD_ATTR_REQ) ;
      if ( nodeptr->AttrReq != ATTR_VALUE ) while ( *p && *p != 32 ) p++ ;
      if ( nodeptr->AttrReq == ATTR_FIXED || nodeptr->AttrReq == ATTR_VALUE )
      {
        if ( *p == 0 ) return(E_DTD_ATTR_FIX) ;
        if ( nodeptr->AttrReq == ATTR_FIXED ) p++ ;
        gotquote = 32 ;
        if ( *p == '"' || *p == '\'' ) gotquote = *p ;
        q = ( gotquote != 32 ) ? p + 1 : p ;
        p++ ;
        while ( *p && *p != gotquote ) p++ ;
        if ( *p == 0 && gotquote != 32 ) return(E_DTD_ATTR_QUOTE) ;
        if ( *p != 0 ) { *p = 0 ; p++ ;}
        nodeptr->NodeData = q ;
      }
    }
    while ( *p && *p == 32 ) p++ ;
    if ( *p == 0 ) break ;
  }

  return(0) ;

}


int XML::DTDNodeAttributeReq( char *reqptr )
{

  char *p ;

  
  p = reqptr ;

  if ( *p == '"' || *p == '\'' ) return(ATTR_VALUE) ;

  if ( strncmp( (const char *)reqptr, ATTR_REQUIRED_STR, LEN_ATTR_REQUIRED_STR ) == 0 )
  {
    p += LEN_ATTR_REQUIRED_STR ;
    return( ((*p != 0 && *p != 32 ) ? E_DTD_ATTR_REQ : ATTR_REQUIRED )) ;
  }

  if ( strncmp( (const char *)reqptr, ATTR_IMPLIED_STR, LEN_ATTR_IMPLIED_STR ) == 0 )
  {
    p += LEN_ATTR_IMPLIED_STR ;
    return( ((*p != 0 && *p != 32 ) ? E_DTD_ATTR_REQ : ATTR_IMPLIED )) ;
  }

  if ( strncmp( (const char *)reqptr, ATTR_FIXED_STR, LEN_ATTR_FIXED_STR ) == 0 )
  {
    p += LEN_ATTR_FIXED_STR ;
    return( ((*p != 0 && *p != 32 ) ? E_DTD_ATTR_REQ : ATTR_FIXED )) ;
  }

  return(E_DTD_ATTR_REQ) ;

}


int XML::DTDNodeAttributeType( char *naptr, int *toktype )
{


  *toktype = 0 ;
  if ( naptr[0] == '(' )
    return(ATTRIBUTE_TYPE_ENUM) ;

  if ( strncmp( (const char *)naptr, ATTR_STRING, LEN_ATTR_STRING ) == 0 )
    return(ATTRIBUTE_TYPE_STRING) ;

  if ( strncmp( (const char *)naptr, TOKEN_ID_STR, LEN_TOKEN_ID_STR ) == 0 )
  {
    *toktype = TOKEN_ID ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  if ( strncmp( (const char *)naptr, TOKEN_IDREF_STR, LEN_TOKEN_IDREF_STR ) == 0 )
  {
    *toktype = TOKEN_IDREF ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  if ( strncmp( (const char *)naptr, TOKEN_IDREFS_STR, LEN_TOKEN_IDREFS_STR ) == 0 )
  {
    *toktype = TOKEN_IDREFS ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  if ( strncmp( (const char *)naptr, TOKEN_ENTITY_STR, LEN_TOKEN_ENTITY_STR ) == 0 )
  {
    *toktype = TOKEN_ENTITY ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  if ( strncmp( (const char *)naptr, TOKEN_ENTITIES_STR, LEN_TOKEN_ENTITIES_STR ) == 0 )
  {
    *toktype = TOKEN_ENTITIES ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  if ( strncmp( (const char *)naptr, TOKEN_NMTOKEN_STR, LEN_TOKEN_NMTOKEN_STR ) == 0 )
  {
    *toktype = TOKEN_NMTOKEN ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  if ( strncmp( (const char *)naptr, TOKEN_NMTOKENS_STR, LEN_TOKEN_NMTOKENS_STR ) == 0 )
  {
    *toktype = TOKEN_NMTOKENS ;
    return(ATTRIBUTE_TYPE_TOKEN) ;
  }

  return(E_DTD_ATTR_TYPE) ;

}


int XML::ResolveAttlistEntities( struct _BaseDTD *AttlistPtr )
{

  char *p, *p1, *q, *r, *s, *t, *u, *v ;
  char *epd ;
  int free_it = 0 ;


  epd = v = AttlistPtr->NodeData ;
  for ( ; ; )
  {
    p = strstr( (const char *)epd, "%" ) ;
    p1 = strstr( (const char *)epd, "&" ) ;
    if ( p == NULL && p1 == NULL ) break ;
    q = strstr((const char *)epd, ";" ) ;
    if ( q == NULL )
      return(E_DTD_ENTITY_SUBS) ;
    else
    {
      if ( p1 != NULL )
      {
        if ( p1 < p ) p = p1 ;
      }
      if ( p == NULL ) p = p1 ;
      r = (char *)malloc( q - p + 2 ) ;
      t = r ;
      while ( p <= q ) *t++ = *p++ ;
      *t = 0 ; 
      u = strdup( (r+1) ) ;
      u[strlen(r) - 2] = 0 ;
      s = ResolveEntity( (const char *)u ) ;
      if ( s == NULL )
      {
        if ( free_it == 1 ) free(epd) ;
        free(u) ;
        free(r) ;
        return(E_DTD_ENTITY_BAD) ;
      }
      v = Replace( epd, r, s ) ;
      AttlistPtr->FreeAtt = 1 ;
      if ( free_it == 1 ) free(epd) ;
      epd = v ;
      free_it = 1 ;
      free(u) ;
      free(r) ;
    }
  }

  p = q = epd ;
  while ( *q )
  {
    if ( *q < 32 || *q > 126 ) *q = 32 ;
    q++ ;
  }
  q = p ;
  while( *q )
  {
    if ( *q != 32 )
    {
      *p++ = *q++ ;
    }
    else
    {
      *p++ = *q++ ;
      while ( *q == 32 ) q++ ;
    }
  }
  if ( *p == 32 ) p-- ;
  *p = 0 ;
  AttlistPtr->Attributes = epd ;

  return(0) ;

}


int XML::Parse( void ) { return( ParseXML( NULL ) ) ; }


int XML::ParseWithDTD( XML *dtdptr )
{

  if ( ! SYS_DTDREADY( dtdptr->GetSystemState()) ) return( E_XML_NODTD ) ;
  return( ParseXML( dtdptr ) ) ;

}


int XML::ParseXML( XML *dtdptr )
{

  char *p, *q, *p1, *p2, *s ;
  int mode, r ;
  int ct, ns, ic ;


  if ( SYS_XMLREADY(SystemState) ) return(E_XML_ALREADY) ;
  if ( ! SYS_XMLLOADED ) return(E_XML_NOXML) ;

  p = SkipWhiteSpace( XMLData ) ;
  if ( *p == '<' && *(p+1) == '?' )
  {
    p = SkipWhiteSpace( p+2 ) ;
    if ( (*p & ~32)      == 'X' &&
         (*(p+1) & ~32 ) == 'M' &&
         (*(p+2) & ~32 ) == 'L' &&
          *(p+3)         == 32
       )
    {
      p = SkipWhiteSpace( p+4 ) ;
      XMLVersion = p ;
      while ( *p && *p != '?' ) p++ ;
      if ( *p != '?' ) XMLVersion = p ;
      *p++ = 0 ;
    }
    else
    {
      *p = 0 ;
      XMLVersion = p++ ;
    }
    while( *p && *p != '>' ) p++ ;
    p = SkipWhiteSpace( p+1 ) ;
  }

  if ( *p == '<' && *(p+1) == '!' )
  {
    p = SkipWhiteSpace( p+2 ) ;
    if ( (*p & ~32)      == 'D' &&
         (*(p+1) & ~32 ) == 'O' &&
         (*(p+2) & ~32 ) == 'C' &&
         (*(p+3) & ~32 ) == 'T' &&
         (*(p+4) & ~32 ) == 'Y' &&
         (*(p+5) & ~32 ) == 'P' &&
         (*(p+6) & ~32 ) == 'E' &&
          *(p+7)         == 32
       )
    {
      p = SkipWhiteSpace( p+8 ) ;
      DocType = p ;
      while( *p && *p != '>' ) p++ ;
      if ( *p ) *p = 0 ;
      p = SkipWhiteSpace( p+1 ) ;
    }
  }

  NodeLevel = 0 ;
  p = SkipWhiteSpace( p ) ;
  for(  mode = 0 ;  ; )
  {

    switch( mode )
    {
      case 0:
        if ( *p != '<' ) 
        {

/*
          if ( XMLDataLength - (p - XMLData) < 5 )
            MessageBox(0, "At the end!!","Parse (End)",MB_OK ) ;
          if ( p - XMLData > 40 )
            MessageBox( 0, p - 40, "Parse 40", MB_OK);
          else
            if ( p -XMLData > 20 )
              MessageBox( 0, p - 20, "Parse 20", MB_OK);
            else
              MessageBox( 0, p - 5, "Parse 5", MB_OK ) ;
*/
          if ( CurrentNode->Index == 0 )
            break ;
          return( E_XML_INVALID_CHAR ) ;
        }
        if ( CurrentNode->Type == NODE_FRAG ) *p = 0 ;
        if ( *(p+1) == '/' )
        {
          CurrentNode = CurrentNode->Parent ;
          mode = 2 ;
        }
        else
        {
          p = SkipWhiteSpace( p+1 ) ;
          p = SkipComments( p, &ic ) ;
          if ( ic == 1 ) continue ;
          q = p ;
          p = FindClosing( p ) ;
          if ( *p != '>' ) 
            return( E_XML_INVALID_TAG ) ;
          *p++ = 0 ;
          NodeLevel++ ;
          ct = -1 ;
          if ( ( r = AddNode( q, NODE_BRANCH, dtdptr ) ) != 0 ) return(r) ;
          if ( ( s = CheckSpecialNodes( p - 1, q, CurrentNode->Type)) != NULL )
          {
            p = s ;
            mode = 0 ;
            continue ;
          }
          if ( *(p-2) == '/' || ( dtdptr != NULL && (ct = dtdptr->GetContentType()) == ELEMENT_CONTENT_EMPTY ) || CurrentNode->Type == NODE_CDATA )
          {
             CurrentNode->Value = p - 1 ;
             if (ct != ELEMENT_CONTENT_EMPTY )
             {
               CurrentNode->Type = NODE_LEAF ;
               CurrentNode->Level = NodeLevel ;
               if ( CurrentNode->Parent == CurrentNode->PreviousSibbling )
               {
                 NodeLevel-- ;
                 CurrentNode->Level = NodeLevel ;
                 CurrentNode->Parent = CurrentNode->Parent->Parent ;
               }
             }
             else
             {
               if ( CurrentNode->Parent->Type == NODE_EMPTY || CurrentNode->Parent->Type == NODE_CDATA )
               {
                 CurrentNode->Parent->NextSibbling = CurrentNode ;
                 CurrentNode->PreviousSibbling = CurrentNode->Parent ;
                 CurrentNode->Parent = CurrentNode->Parent->Parent ;
                 NodeLevel-- ;
                 CurrentNode->Level = NodeLevel ;
               }
               CurrentNode->Type = NODE_EMPTY ;
               if ( CurrentNode->Parent != NULL && CurrentNode->PreviousSibbling != NULL )
               {
                 if ( CurrentNode->Parent == CurrentNode->PreviousSibbling->Parent )
                 {
                   CurrentNode->PreviousSibbling->NextSibbling = CurrentNode ;
                 }
               }
             }
          }
          else
          {
            if ( CurrentNode->Parent != NULL && CurrentNode->PreviousSibbling != NULL )
            {
              if ( CurrentNode->Parent == CurrentNode->PreviousSibbling->Parent )
              {
                CurrentNode->Type = NODE_BRANCH ;
                CurrentNode->PreviousSibbling->NextSibbling = CurrentNode ;
              }
            }
          }
          if ( CurrentNode->Type == NODE_EMPTY || CurrentNode->Type == NODE_CDATA )
            mode = 0 ;
          else
            mode = 1 ;
        }
        break ;


      case 1:
        p = SkipWhiteSpace( p ) ;
        if ( *p != '<' )
        {
//         if ( CurrentNode->Parent->Level < CurrentNode->Level && CurrentNode->Parent->Type == NODE_BRANCH )
         if ( CurrentNode->Parent->Level < CurrentNode->Level && (CurrentNode->Parent->FirstChild != NULL || CurrentNode->Parent->Type == NODE_EMPTY || CurrentNode->Parent->Type == NODE_CDATA ))
          {
            CurrentNode->Type = NODE_LEAF ;
            NodeLevel-- ;
            CurrentNode->Level = NodeLevel ;
            if ( CurrentNode->PreviousSibbling != NULL )
              CurrentNode->PreviousSibbling->NextSibbling = CurrentNode ;
            if ( CurrentNode->Parent->Type == NODE_EMPTY || CurrentNode->Parent->Type == NODE_CDATA )
            {
              CurrentNode->Parent->NextSibbling = CurrentNode ;
              CurrentNode->PreviousSibbling = CurrentNode->Parent ;
              CurrentNode->Parent = CurrentNode->Parent->Parent ;
            }
            mode = 3 ;
            break ;
          }
          else 
            return( E_XML_INVALID_CHAR2 ) ;
        }
        if ( *(p+1) == '/' )
        {
           if ( CurrentNode->PreviousSibbling != NULL )
           {
             if ( CurrentNode->PreviousSibbling->FirstChild == NULL || CurrentNode->FirstChild == NULL)
             {
               if ( strncmp((const char *)q,(const char *)(p+2),strlen((const char *)q)) == 0 && *(p + 2 + strlen((const char *)q)) == '>')
                 CurrentNode = CurrentNode ;
               else
                 CurrentNode = CurrentNode->Parent ;
             }
           }
           else
             CurrentNode = CurrentNode->Parent ;
           NodeLevel = CurrentNode->Level ;
           mode = 2 ;
        }
        else
        {
  /*        nodestate = 0 ; */
          if ( *p == '<' )
          {
            p = SkipComments( p+1, &ic ) ;
            if ( ic == 1 ) 
              continue ;
            else
              p-- ;
          }
          if ( CurrentNode->PreviousSibbling != NULL )
          {
            NodeLevel = CurrentNode->PreviousSibbling->Level ;
            CurrentNode->Level = NodeLevel ;
            CurrentNode->Parent = CurrentNode->PreviousSibbling->Parent ;
          }
          p = SkipWhiteSpace( p+1 ) ;
          if ( *p == 0 ) 
            return( E_XML_INVALID_TAG ) ;
          q = p ;
          p = FindClosing( p ) ;
          if ( *p != '>' ) 
            return( E_XML_INVALID_TAG ) ;
          ns = 0 ;
          if ( *(p-1) == '/' )
          {
            ns = 1 ;
            *(p-1) = 0 ;
          }
          *p++ = 0 ;
          ct = -1 ;
          if ( ( r = AddNode( q, NODE_LEAF, dtdptr ) ) != 0 ) return(r) ;
          if ( ( s = CheckSpecialNodes( p - 1, q, CurrentNode->Type)) != NULL )
          {
            p = s ;
            mode = 0 ;
//            mode = 1 ;
            continue ;
          }
          if ( CurrentNode->Level == CurrentNode->Parent->Level )
          {
             CurrentNode->Level++ ;
 //            NodeLevel++ ;
          }
          if ( CurrentNode->Type == NODE_CDATA || ns == 1 || ( dtdptr != NULL && (ct = dtdptr->GetContentType()) == ELEMENT_CONTENT_EMPTY ) )
          {
            CurrentNode->Value = p - 1 ;
            if ( ct == ELEMENT_CONTENT_EMPTY )
            {
              CurrentNode->Type = NODE_EMPTY ;
              mode = 0 ;
            }
          }
          else
            mode = 3 ;
        }
        break ;
      
      case 2:
        if ( CurrentNode->Type != NODE_COMMENT )
        {
          q = p + 2 ;
          p = FindClosing( p+2 ) ;
          *p++ = 0 ;
          if ( stricmp( (const char *)CurrentNode->Name, (const char *)q ) != 0 )
            return( E_XML_MATCH ) ;
          p = SkipWhiteSpace( p ) ;
          NodeLevel-- ;
        }
        mode = 0 ;
        break ;

      case 3:
        if ( CurrentNode->PreviousSibbling != NULL ) CurrentNode->Level = CurrentNode->PreviousSibbling->Level ;
        p = SkipWhiteSpace( p ) ;
        if ( *p != '<' )
        {
           q = p1 = p ;
           p = FindEndTag( p ) ;
           if ( *p == '<' && *(p+1) == '/' )
           {
             *p = 0 ;
             p2 = CurrentNode->Value ;
             CurrentNode->Value = q ;
             q = p + 2 ;
             p = FindClosing( p+1 ) ;
             *p++ = 0 ;
             if ( stricmp( (const char *)CurrentNode->Name, (const char *)q ) != 0 )
             {
               if ( stricmp( (const char *)CurrentNode->Parent->Name, (const char *)q ) == 0 )
               {
                 CurrentNode->Value = p2 ;
                 if ( ( r = AddNode( NULL, NODE_FRAG, dtdptr ) ) != 0 ) return(r) ;
                 CurrentNode->Value = p1 ;
                 *(q-2) = 0 ;
                 CurrentNode = CurrentNode->Parent ;
               }
               else             
                 return( E_XML_MATCH ) ;
             }
             mode = 1 ;
           }
           else
           {
             if ( *p == '<' )
             {
               p = SkipComments( p, &ic ) ;
               if ( ic == 1 ) continue ;
               CurrentNode->Type = NODE_BRANCH ;
               if ( ( r = AddNode( NULL, NODE_FRAG, dtdptr ) ) != 0 ) return(r) ;
               CurrentNode->Value = q ;
               mode = 0 ;
             }

           }
        }
        else
        {
          if ( *(p+1) == '/' )
          {
             q = p + 2 ;
             p = FindClosing( p+1 ) ;
             *p++ = 0 ;
             if ( stricmp( (const char *)CurrentNode->Name, (const char *)q ) != 0 )
               return( E_XML_MATCH ) ;
             CurrentNode->Value = p - 1 ; 
             mode = 1 ;
          }
          else
          {
            p = SkipComments( p, &ic ) ;
            if ( ic == 1 ) continue ;
            CurrentNode->Type = NODE_BRANCH ;
            NodeLevel++ ;
            CurrentNode->Level = NodeLevel ;
            CurrentNode->FirstChild = NULL ;
            CurrentNode->NextSibbling = NULL ;
            if ( CurrentNode->Parent->FirstChild == NULL )
              CurrentNode->PreviousSibbling = NULL ;
            mode = 1 ;
          }
        }
        break ;

      default:
        return( E_XML_MODE ) ;
        break ;
    }
    p = SkipWhiteSpace( p ) ;
    if ( XMLDataLength - (p - XMLData)  < 5 || *p == 0 || CurrentNode == NULL )
      break ;
  }

  if ( CurrentNode->Parent != NULL )
    return( E_XML_MISSING ) ;
  else
  {
    Unescape() ;
    CurrentNode = &Top ;
    SystemState |= SYS_STATE_XMLPARSED ;
    return(0) ;
  }

}


char *XML::SkipComments( char *p, int *IsComment )
{

  *IsComment = 0 ;
  if ( *p == '!' && *(p+1) == '-' && *(p+2) == '-' )
  {
    *IsComment = 1 ;
    while( *p && strncmp((const char *)p, E_COMMENT, E_LEN_COMMENT ) != 0 ) p++ ;
    p = p + E_LEN_COMMENT ;
  }

  return(SkipWhiteSpace(p)) ;

}


char * XML::CheckSpecialNodes( char *gt, char *p, int NodeType )
{

  switch( NodeType )
  {
    case NODE_CDATA:
      *gt = '>' ;
      CurrentNode->Value = p + 8 ;
      while( *p && strncmp((const char *)p, CDATA_END, LEN_CDATA_END ) != 0 ) p++ ;
      if ( *p == 0 ) return(p) ;
      *p = 0 ;
      p = p + LEN_CDATA_END ;
      break ;

/*
    case NODE_COMMENT:
      NodeLevel-- ;
      CurrentNode->Level = NodeLevel ;
      *gt = '>' ;
      CurrentNode->Value = p + 2 ;
      while( *p && strncmp((const char *)p, E_COMMENT, E_LEN_COMMENT ) != 0 ) p++ ;
      if ( *p == 0 ) return(p) ;
      *p = 0 ;
      p = p + E_LEN_COMMENT ;
      break ;
*/
    default:
      return(NULL) ;
      break ;
  }

  p = SkipWhiteSpace(p) ;
  *gt = 0 ;

  return(p) ;

}


void XML::Unescape( void )
{

  long i ;


  if ( Top.NextIndex != NULL )
  {
    CurrentNode = &Top ;
    for ( i = 0 ; i <= Index ; i++ )
    {
      if ( CurrentNode->Value != NULL )
      {
         if ( strlen( (const char *)CurrentNode->Value) > 0 )
           ReplaceEscapes() ;
      }
      CurrentNode = CurrentNode->NextIndex ;
    }
  }

}


void XML::ReplaceEscapes( void )
{

  char *p, *q ;
  int i ;


  p = q = CurrentNode->Value ;
  while ( *q )
  {
    for ( i = 0 ; i < NUMESCAPES ; i++ )
    { 
      if ( strnicmp( (const char *)q, esc[i].eSeq, esc[i].eLen ) == 0 )
      {
        *p++ = esc[i].eChar ;
        q += esc[i].eLen ;
        break ;
      }
    }
    *p++ = *q++ ;
  }
  *p = 0 ;

}


int XML::AddNode( char *p, int NodeType, XML *dtdptr )
{

  struct _Node *nodeptr ;
  int ret = 0 ;


  if ( CurrentNode->Name != NULL || CurrentNode->Type == NODE_FRAG || CurrentNode->Type == NODE_COMMENT)
  {
    if ( p != NULL )
    {
      if ( strncmp( (const char *)p, CDATA_XSTART, strlen(CDATA_XSTART) ) == 0 )
        NodeType = NODE_CDATA ;
    }
    Index++ ;
    nodeptr = new struct _Node ;
    nodeptr->Index = Index ;
    nodeptr->NextIndex = NULL ;
    nodeptr->PrevIndex = LastNode ;
    LastNode->NextIndex = nodeptr ;
    LastNode = nodeptr ;
    switch ( NodeType )
    {
      case NODE_BRANCH:
        if ( CurrentNode->Level == NodeLevel || CurrentNode->Type == NODE_FRAG )
        {
          nodeptr->Parent = CurrentNode->Parent ;
          nodeptr->PreviousSibbling = CurrentNode ;
//          if ( CurrentNode->Type == NODE_FRAG )
//            NodeLevel-- ;
          if ( CurrentNode->PreviousSibbling != NULL )
            CurrentNode->PreviousSibbling->NextSibbling = CurrentNode ;
        }
        else
        {
          nodeptr->Parent = CurrentNode ;
          nodeptr->PreviousSibbling = NULL ;
        }
        nodeptr->FirstChild = NULL ;
        nodeptr->NextSibbling = NULL ;
        if ( CurrentNode->PreviousSibbling == NULL )
        {
          nodeptr->PreviousSibbling = CurrentNode ;
          CurrentNode->NextSibbling = nodeptr ;
          nodeptr->NextSibbling = NULL ;
        }
        break ;

      case NODE_COMMENT:
        nodeptr->Parent = CurrentNode->Parent ;
        nodeptr->NextSibbling = NULL ;
        nodeptr->FirstChild = NULL ;
        nodeptr->PreviousSibbling = NULL ;
        if ( CurrentNode->NextSibbling == NULL )
        {
          CurrentNode->NextSibbling = nodeptr ;
          nodeptr->PreviousSibbling = CurrentNode ;
        }      
        break ;

      case NODE_CDATA:
      case NODE_FRAG:
      case NODE_LEAF:
        if ( CurrentNode->Type == NODE_BRANCH )
        {
          if ( NodeType == NODE_FRAG ) 
            NodeLevel++ ;
          nodeptr->Parent = CurrentNode ;
          nodeptr->NextSibbling = NULL ;
          nodeptr->FirstChild = NULL ;
          nodeptr->PreviousSibbling = NULL ;
          if ( CurrentNode->FirstChild == NULL )
          {
            CurrentNode->FirstChild = nodeptr ;
            if ( CurrentNode->Level == NodeLevel && NodeType == NODE_FRAG  ) NodeLevel++ ;
          }
        }
        else
        {
          nodeptr->Parent = CurrentNode->Parent ;
          nodeptr->NextSibbling = NULL ;
          nodeptr->FirstChild = NULL ;
          nodeptr->PreviousSibbling = NULL ;
          if ( CurrentNode->NextSibbling == NULL )
          {
            CurrentNode->NextSibbling = nodeptr ;
            nodeptr->PreviousSibbling = CurrentNode ;
          }      
        }
        break ;

    }
    CurrentNode = nodeptr ;
  }
  CurrentNode->Level = NodeLevel ;
  CurrentNode->Type = NodeType ;
  CurrentNode->AttrList = NULL ;
  CurrentNode->Value = NULL ;
  CurrentNode->Attributes = NULL ;
  CurrentNode->AttrValues = NULL ;
  if ( CurrentNode->PreviousSibbling != NULL )
  {
    if ( CurrentNode->Level == CurrentNode->PreviousSibbling->Level )
      CurrentNode->PreviousSibbling->NextSibbling = CurrentNode ;
  }
  if ( NodeType == NODE_CDATA || NodeType == NODE_COMMENT )
  {
    CurrentNode->Name = NULL ;
    return(0) ;
  }
  if ( NodeType == NODE_FRAG )
  {
    CurrentNode->Name = NULL ;
  }
  else
  {
    CurrentNode->Name = p ;
    while ( *p++ )
    { 
      if ( *p == ' ' )
      {
         *p = 0 ;
         CurrentNode->AttrList = p + 1 ;
         ret = AttributeNodes( p + 1 ) ;
         break ;
      }
    }
 
    if ( dtdptr != NULL && ret == 0 && 
         ElementHasNamespace((const char *)CurrentNode->Name) == 0 )
    {
      if ( dtdptr->FindElement((const char *)CurrentNode->Name ) != 0 )
        return(E_XML_DTDELEMENT) ;
    }
  }

  return(ret) ;

}


int XML::AttributeNodes( char *attstart )
{

  char *p, *q, *r ;
  struct _Node *nodeptr, *prevnode ;


  prevnode = CurrentNode ;
  p = SkipWhiteSpace(attstart) ;
  q = FindClosing( attstart ) ;
  if ( *(q-1) == '/' ) q-- ;

  CurrentNode->AttrValues = new char [q - p + 1 ] ;
  r = CurrentNode->AttrValues ;
  while( p < q ) *r++ = *p++ ;
  *r = 0 ;

  p = q = CurrentNode->AttrValues ;


  while( *p )
  {

//    MessageBox( 0, p, "XMLParse",MB_OK) ;

    while ( *q != '=' && *q != ' ' ) q++ ;
    *q = 0 ;
    Index++ ;
    nodeptr = new struct _Node ;
    nodeptr->Type = NODE_ATTRVAL ;
    nodeptr->Level = CurrentNode->Level ;
    nodeptr->Parent = CurrentNode ;
    nodeptr->Index = Index ;  
    nodeptr->NextIndex = NULL ;
    nodeptr->PrevIndex = LastNode ;
    LastNode->NextIndex = nodeptr ;
    LastNode = nodeptr ;
    nodeptr->Name = p ;
    nodeptr->Attributes = NULL ;
    nodeptr->AttrValues = NULL ;
    nodeptr->FirstChild = NULL ;

    if ( prevnode == CurrentNode )
    {
      CurrentNode->Attributes = nodeptr ;
      nodeptr->PreviousSibbling = NULL ;
    }
    else
    {
      nodeptr->PreviousSibbling = prevnode ;
      prevnode->NextSibbling = nodeptr ;
    }

    nodeptr->NextSibbling = NULL ;

    p = q + 1 ;
    p = SkipWhiteSpace(p) ;
//    if ( *p != '=' ) return(E_XML_BAD_ATTR) ;
//    p = SkipWhiteSpace(p+1) ;
    if ( *p != '"' && *p != '\'' ) 
      return(E_XML_ATTR_VAL) ;
    r = p ;
    p++ ;
    nodeptr->Value = p ;
//    while( *p != '"' ) p++ ;
    while( *p != *r ) p++ ;
    *p++ = 0 ;
    p = SkipWhiteSpace(p) ;
    prevnode = nodeptr ;
    q = p ;
  }

  return(0) ;

}


int XML::ElementHasNamespace( const char *eName )
{

  char *p ;


  p  = strstr(eName, ":") ;
  if ( p != NULL )
    p = p ;
  return( (p == NULL) ? 0 : p - eName ) ;

}


char *XML::SkipWhiteSpace( char *p )
{

  while ( *p && (*p == ' ' || *p == 9 || *p == 10 || *p == 13) ) p++ ;
  return( p ) ;

}


char *XML::FindWhiteSpace( char *p )
{

  while ( *p && (*p != ' ' && *p != 9 && *p != 10 && *p != 13) ) p++ ;
  return( p ) ;

}


char *XML::FindClosing( char *p )
{

  int out, gotquote ;


  out = gotquote = 0 ;
  while( *p && out == 0 )
  {
     switch( *p )
     {
       case 0:
         out = 1 ;
         break ;
       case '>':
         if ( gotquote == 0 ) { out = 1 ; break ; }
       case '"':
         gotquote = ( gotquote == 1 ) ? 0 : 1 ;
       default:
         p++ ;
         break ;
     }
  }
  return( p ) ;

}


char *XML::FindEndTag( char *p )
{

  int out, gotquote ;


  out = gotquote = 0 ;
  while( *p && out == 0 )
  {
     switch( *p )
     {
       case 0:
         out = 1 ;
         break ;
       case '<':
         if ( gotquote == 0 ) { out = 1 ; break ; }
       case '"':
         gotquote = ( gotquote == 1 ) ? 0 : 1 ;
       default:
         p++ ;
         break ;
     }
  }
  return( p ) ;


}


int XML::FindFirstNode( const char *NodeName )
{

  return( FindNode( NodeName, 1 ) ) ;

}


int XML::FindNode( const long NodeIndex )
{

  long i ;


  if ( NodeIndex <= Index )
  {
    CurrentNode = &Top ;
    for ( i = 0 ; i < NodeIndex ; i++ )
      CurrentNode = CurrentNode->NextIndex ;
    return(0) ;
  }
  return(1) ;

}


int XML::FindNode( const char *NodeName, const long Occurence )
{

  int i ;


  CurrentNode = &Top ;
  i = 0 ;
  do
  {
    
    if ( stricmp( NodeName, CurrentNode->Name ) == 0 )
    {
      i++ ;
      if ( i == Occurence ) return(0) ;
    }
    if ( CurrentNode->FirstChild != NULL )
      CurrentNode = CurrentNode->FirstChild ;
    else
      if ( CurrentNode->NextSibbling != NULL ) 
        CurrentNode = CurrentNode->NextSibbling ;
      else
      {
        CurrentNode = CurrentNode->Parent ;
        if ( CurrentNode->NextSibbling != NULL )
          CurrentNode = CurrentNode->NextSibbling ;
        else
        {
          while( CurrentNode->NextSibbling == NULL && CurrentNode->Parent != NULL ) 
            CurrentNode = CurrentNode->Parent ;
          if ( CurrentNode->Parent != NULL ) CurrentNode = CurrentNode->NextSibbling ;
        }
      }
  } while ( CurrentNode->Parent != NULL ) ;

  return( 1 ) ;

}


int XML::FindNode( const int NodeType, const long Occurence )
{

  int i ;


  CurrentNode = &Top ;
  i = 0 ;
  do
  {
    
    if ( NodeType == CurrentNode->Type )
    {
      i++ ;
      if ( i == Occurence ) return(0) ;
    }
    if ( CurrentNode->FirstChild != NULL )
      CurrentNode = CurrentNode->FirstChild ;
    else
      if ( CurrentNode->NextSibbling != NULL ) 
        CurrentNode = CurrentNode->NextSibbling ;
      else
      {
        CurrentNode = CurrentNode->Parent ;
        if ( CurrentNode->NextSibbling != NULL )
          CurrentNode = CurrentNode->NextSibbling ;
        else
        {
          while( CurrentNode->NextSibbling == NULL && CurrentNode->Parent != NULL ) 
            CurrentNode = CurrentNode->Parent ;
          if ( CurrentNode->Parent != NULL ) CurrentNode = CurrentNode->NextSibbling ;
        }
      }
  } while ( CurrentNode->Parent != NULL ) ;

  return( 1 ) ;

}


int XML::FindNode( const char *NodeName, const char *AttrName, const char *AttrValue )
{


  CurrentNode = &Top ;
  do
  {
    
    if ( strcmp( (const char *)CurrentNode->Name, NodeName) == 0 && 
         NodeHasAttribute(AttrName) &&
         strcmp(GetNodeAttrValue(AttrName), AttrValue) == 0
       )
    {
      return(0) ;
    }
    if ( CurrentNode->FirstChild != NULL )
      CurrentNode = CurrentNode->FirstChild ;
    else
      if ( CurrentNode->NextSibbling != NULL ) 
        CurrentNode = CurrentNode->NextSibbling ;
      else
      {
        CurrentNode = CurrentNode->Parent ;
        if ( CurrentNode->NextSibbling != NULL )
          CurrentNode = CurrentNode->NextSibbling ;
        else
        {
          while( CurrentNode->NextSibbling == NULL && CurrentNode->Parent != NULL ) 
            CurrentNode = CurrentNode->Parent ;
          if ( CurrentNode->Parent != NULL ) CurrentNode = CurrentNode->NextSibbling ;
        }
      }
  } while ( CurrentNode->Parent != NULL ) ;

  return( 1 ) ;

}


const char *XML::NodeValue( void )
{

  return( (const char *)CurrentNode->Value ) ;

}


const int XML::NodeType( void )
{

  return( (const int )CurrentNode->Type ) ;

}


const char *XML::NodeAttr( void )
{

  return( (const char *)CurrentNode->AttrList ) ;

}


const char *XML::GetVersion( void )
{

  if ( XMLVersion == NULL )
    return( "" ) ;
  else
    return( (const char *)XMLVersion ) ;

}


const char *XML::GetDocType( void )
{

  if ( DocType == NULL )
    return( "" ) ;
  else
    return( (const char *)DocType ) ;

}


void XML::MoveToTop( void )
{

  CurrentNode = &Top ;

}


const char *XML::GetNodeAttrValue( const char *at )
{

  struct _Node *tmpnode ;

  if ( CurrentNode->Attributes != NULL )
  {
     tmpnode = CurrentNode->Attributes ;
     for ( ; ; )
     {
       if ( stricmp( at, (const char *)tmpnode->Name ) == 0 )
         return( (const char *)tmpnode->Value ) ;
       if ( tmpnode->NextSibbling != NULL )
         tmpnode = tmpnode->NextSibbling ;
       else
         break ;
     }
  }

  return( (const char *)NULL ) ;

}


int XML::NodeHasAttribute(const char *AttrName)
{
  struct _Node *tmpnode ;

  if ( CurrentNode->Attributes != NULL )
  {
     tmpnode = CurrentNode->Attributes ;
     for ( ; ; )
     {
       if ( stricmp( AttrName, (const char *)tmpnode->Name ) == 0 )
         return( 1 ) ;
       if ( tmpnode->NextSibbling != NULL )
         tmpnode = tmpnode->NextSibbling ;
       else
         break ;
     }
  }

  return( 0 ) ;

}


int XML::DocNodeHasAttribute(const char *AttrName)
{
  struct _Node *tmpnode ;

  if ( CurrentDocNode->Attributes != NULL )
  {
     tmpnode = CurrentDocNode->Attributes ;
     for ( ; ; )
     {
       if ( stricmp( AttrName, (const char *)tmpnode->Name ) == 0 )
         return( 1 ) ;
       if ( tmpnode->NextSibbling != NULL )
         tmpnode = tmpnode->NextSibbling ;
       else
         break ;
     }
  }

  return( 0 ) ;

}



const char *XML::GetNodeName()
{

  return((const char *)CurrentNode->Name ) ;

}


int XML::MoveToFirstChild( void )
{

  if ( CurrentNode->FirstChild != NULL ) 
  {
    CurrentNode = CurrentNode->FirstChild ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::MoveToNextSibling( void )
{

  if ( CurrentNode->NextSibbling != NULL ) 
  {
    CurrentNode = CurrentNode->NextSibbling ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::MoveToParent( void )
{

  if ( CurrentNode->Parent != NULL ) 
  {
    CurrentNode = CurrentNode->Parent ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::MoveToPrevSibling( void )
{

  if ( CurrentNode->PreviousSibbling != NULL ) 
  {
    CurrentNode = CurrentNode->PreviousSibbling ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::NodeHasChild( void ){return( (CurrentNode->FirstChild != NULL) ? 1 : 0 ) ; }


int XML::NodeHasSibling( void ){return( (CurrentNode->NextSibbling != NULL) ? 1 : 0 ) ; }


int XML::NodeHasParent( void ){return( (CurrentNode->Parent != NULL) ? 1 : 0 ) ; }


long XML::GetNodeIndex( void ){ return( CurrentNode->Index ) ; }


const char *XML::GetParentName( void )
{

  if ( CurrentNode->Parent != NULL )
    return((const char *)CurrentNode->Parent->Name ) ;
  else
    return((const char *)"") ;

}


long XML::GetParentIndex( void )
{

  if ( CurrentNode->Parent != NULL )
    return( CurrentNode->Parent->Index ) ;
  else
    return(-1) ;

}


char *XML::Replace( char *src, char *item, char *repl )
{

  int n, l1, l2, l3 ;
  char *p, *q, *dest ;


  n = 0 ;

    l1 = strlen( (const char *)src ) ;  // Original string
    l2 = strlen( (const char *)item ) ;  // Thing to replace
    l3 = strlen( (const char *)repl ) ;  // Thing to substitute
    p = src ;
    while( *p )
    {
       if ( strncmp( (const char *)p , item, l2 ) == 0 ) 
       {
         n++ ;
         p += l2 ;
       }
       else p++ ;
    }
    dest = (char *)malloc( l1 + ( l3 - l2 ) * n + 1 ) ;
    p = src ;
    q = dest ;
    while( *p )
    {
       if ( strncmp( (const char *)p , item, l2 ) == 0 ) 
       {
         p += l2 ;
         strcpy( q, (const char *)repl ) ;
         q += l3 ;
       }
       else *q++ = *p++ ;
    }

    *q = 0 ;
 

  return(dest) ;

}


unsigned long XML::GetSystemState( void ) { return( SystemState ) ; }


int XML::FindElement( const char *ElementName )
{

  if ( ! SYS_DTDREADY(SystemState) ) return(1) ;

  for( CurrentDTDNode = &DTD ; ; )
  {
    if ( CurrentDTDNode->NextNode != NULL )
    {
      CurrentDTDNode = CurrentDTDNode->NextNode ;
      if ( CurrentDTDNode->Type == NODE_ELEMENT &&
        strcmp( ElementName, (const char *)CurrentDTDNode->Name ) == 0 ) return(0) ;
    }
    else
      break ;
  }
  return(1) ;
  
}


int XML::ValidateXMLvsDTD( XML *xml )
{

  int ret ;


  if ( ! SYS_DTDREADY(SystemState) ) return(E_DTD_READY) ;
  if ( ! SYS_XMLREADY(xml->GetSystemState() ) ) return(E_XML_READY) ;
  if ( (ret = ValidateXMLElements( xml )) != 0 ) 
    return(ret) ;

  return(0) ;

}


int XML::ValidateDocvsDTD( XML *xml )
{

  int ret ;

  if ( xml != NULL )
  {
    if ( ! SYS_DTDREADY(SystemState) ) return(E_DTD_READY) ;
    if ( (ret = ValidateDocElements( xml )) != 0 ) 
      return(ret) ;
  }

  return(0) ;

}


int XML::ValidateNodeAttributes( struct _BaseDTD *dtdnode, XML *xml )
{

  const char *cp ;
  struct _BaseDTD *attrenum ;
  int foundenum ;


  if ( dtdnode->FirstAttr == NULL )
  {
    if ( xml->NodeAttr() == NULL ) return(0) ;
    if ( strlen( xml->NodeAttr() ) == 0 ) return(0) ;
    return(E_XML_HASATTR) ;
  }
  else
  {
    dtdnode = dtdnode->FirstAttr ;
    for ( ; ; )
    {
      if ( (cp = xml->GetNodeAttrValue( (const char *)dtdnode->Name )) == NULL )
        return(E_XML_ATTRNAME) ;
      switch( dtdnode->AttrType )
      {
        case ATTRIBUTE_TYPE_STRING:
        case ATTRIBUTE_TYPE_TOKEN:
          switch( dtdnode->AttrReq )
          {
            case ATTR_REQUIRED:
              if ( cp == NULL || strlen(cp) == 0 )
                return(E_XML_ATTRREQ) ;
              break ;

            case ATTR_FIXED:
            case ATTR_VALUE:
              if ( cp != NULL )
              {
                if ( strlen(cp) == 0 )
                  return(E_XML_ATTRFIXVAL) ;
                else
                {
                  if ( strcmp( (const char *)dtdnode->NodeData, cp ) != 0 )
                    return(E_XML_ATTRBADVAL) ;
                }
              }
              else
                return(E_XML_ATTRFIXVAL) ;
              break ;
          }
          break ;

        case ATTRIBUTE_TYPE_ENUM:
          if ( dtdnode->AttrReq == ATTR_REQUIRED )
          {
              if ( cp == NULL || strlen(cp) == 0 )
                return(E_XML_ATTRREQ) ;
          }
          if ( dtdnode->NodeData != NULL && strcmp( (const char *)dtdnode->NodeData, (const char *)cp) == 0 )
            break ;
          else
          {
            attrenum = dtdnode->FirstChild ;
            for( foundenum = 0 ; ; )
            {
              if ( strcmp( (const char *)cp, (const char *)attrenum->NodeData ) == 0 )
              {
                foundenum = 1 ;
                break ;
              }
              if ( attrenum->NextNode != NULL )
                attrenum = attrenum->NextNode ;
              else
                break ;
            }
            if ( foundenum == 0 ) return(E_XML_ATTRBADENUM) ;
          }
          /* Need code here to handle checking enumeration attribute types */

          break ;
      }
      if ( dtdnode->NextNode == NULL )
        break ;
      else
        dtdnode = dtdnode->NextNode ;
    }
  }

  return(0) ;

}


int XML::DocValidateNodeAttributes( struct _BaseDTD *dtdnode, XML *xml )
{

  const char *cp ;
  struct _BaseDTD *attrenum ;
  int foundenum ;


  if ( dtdnode->FirstAttr == NULL )
  {
    return( ( xml->DocNodeHasAttr() == 0 ) ? 0 : E_DOC_HASATTR) ;
  }
  else
  {
    dtdnode = dtdnode->FirstAttr ;
    for ( ; ; )
    {
      if ( (cp = xml->DocGetNodeAttrValue( (const char *)dtdnode->Name )) == NULL )
      {
        if (dtdnode->AttrReq == ATTR_REQUIRED )
          return(E_DOC_ATTRREQ) ;
        else
          return(E_DOC_ATTRNAME) ;
      }
      switch( dtdnode->AttrType )
      {
        case ATTRIBUTE_TYPE_STRING:
        case ATTRIBUTE_TYPE_TOKEN:
          switch( dtdnode->AttrReq )
          {
            case ATTR_REQUIRED:
              if ( cp == NULL || strlen(cp) == 0 )
                return(E_DOC_ATTRREQ) ;
              break ;

            case ATTR_FIXED:
            case ATTR_VALUE:
              if ( cp != NULL )
              {
                if ( strlen(cp) == 0 )
                  return(E_DOC_ATTRFIXVAL) ;
                else
                {
                  if ( strcmp( (const char *)dtdnode->NodeData, cp ) != 0 )
                    return(E_DOC_ATTRBADVAL) ;
                }
              }
              else
                return(E_DOC_ATTRFIXVAL) ;
              break ;
          }
          break ;

        case ATTRIBUTE_TYPE_ENUM:
          if ( dtdnode->AttrReq == ATTR_REQUIRED )
          {
              if ( cp == NULL || strlen(cp) == 0 )
                return(E_DOC_ATTRREQ) ;
          }
          if ( dtdnode->NodeData != NULL && strcmp( (const char *)dtdnode->NodeData, (const char *)cp) == 0 )
            break ;
          else
          {
            attrenum = dtdnode->FirstChild ;
            for( foundenum = 0 ; ; )
            {
              if ( strcmp( (const char *)cp, (const char *)attrenum->NodeData ) == 0 )
              {
                foundenum = 1 ;
                break ;
              }
              if ( attrenum->NextNode != NULL )
                attrenum = attrenum->NextNode ;
              else
                break ;
            }
            if ( foundenum == 0 ) return(E_DOC_ATTRBADENUM) ;
          }
          /* Need code here to handle checking enumeration attribute types */

          break ;
      }
      if ( dtdnode->NextNode == NULL )
        break ;
      else
        dtdnode = dtdnode->NextNode ;
    }
  }

  return(0) ;

}


int XML::ValidateXMLElements( XML *xml )
{

  int ret ;


    xml->MoveToTop() ;
    if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
    do
    {
      if ( xml->NodeHasChild() == 1 )
      {
        xml->MoveToFirstChild() ;
        if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
      }
      else
      {
        if ( xml->NodeHasSibling() == 1 )
        {
          xml->MoveToNextSibling() ;
          if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
        }
        else
        {
          xml->MoveToParent() ;
          if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
          if ( xml->NodeHasSibling() == 1 )
          {
            xml->MoveToNextSibling() ;
            if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
          }
          else
          {
            while ( xml->NodeHasSibling() == 0 && xml->NodeHasParent() == 1 )
            {
              xml->MoveToParent() ;
              if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
            }
            if ( xml->NodeHasParent() == 1 && xml->NodeHasSibling() == 1 )
            {
              xml->MoveToNextSibling() ;
              if ( (ret = DTDCheckXMLNode(xml)) != 0 ) return(ret) ;
            }
          }
        }
      }
    }
    while ( xml->NodeHasParent() == 1 ) ;

    return(0) ;

}


int XML::ValidateDocElements( XML *xml )
{

  int ret ;


    xml->DocMoveToTop() ;
    if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
    do
    {
      if ( xml->DocNodeHasChild() == 1 )
      {
        xml->DocMoveToFirstChild() ;
        if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
      }
      else
      {
        if ( xml->DocNodeHasSibling() == 1 )
        {
          xml->DocMoveToNextSibling() ;
          if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
        }
        else
        {
          xml->DocMoveToParent() ;
          if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
          if ( xml->NodeHasSibling() == 1 )
          {
            xml->DocMoveToNextSibling() ;
            if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
          }
          else
          {
            while ( xml->DocNodeHasSibling() == 0 && xml->DocNodeHasParent() == 1 )
            {
              xml->DocMoveToParent() ;
              if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
            }
            if ( xml->DocNodeHasParent() == 1 && xml->DocNodeHasSibling() == 1 )
            {
              xml->DocMoveToNextSibling() ;
              if ( (ret = DTDCheckDocNode(xml)) != 0 ) return(ret) ;
            }
          }
        }
      }
    }
    while ( xml->DocNodeHasParent() == 1 ) ;

    return(0) ;

}


int XML::DTDCheckXMLNode( XML *xml )
{

  struct _BaseDTD *nodeptr, *parent, *nextsib, *prevsib ;
  long nIndex, pIndex, nsIndex, psIndex ;
  int ret ;


  if ( xml->NodeType() == NODE_FRAG )
  {
    nodeptr = NULL ;
  }
  else
  {
    if ( FindElement( xml->GetNodeName() ) != 0 ) 
      return(E_XML_ELEMENT) ;
    nodeptr = CurrentDTDNode ;
    if ( (ret = ValidateNodeAttributes( nodeptr, xml )) != 0 ) return(ret) ;
    nodeptr = CurrentDTDNode ;
  }

  nIndex = xml->GetNodeIndex() ;
  parent = nextsib = prevsib = NULL ;

  if ( xml->NodeHasParent() == 1 )
  {
    xml->MoveToParent() ;
    pIndex = xml->GetNodeIndex() ;
    if ( FindElement( xml->GetNodeName() ) == 0 )
      parent = CurrentDTDNode ;
    xml->FindNode( (const long)nIndex ) ;
  }
  else
    pIndex = -1 ;

  if ( xml->MoveToNextSibling() == 0 )
  {
    nsIndex = xml->GetNodeIndex() ;
    if ( xml->NodeType() == NODE_FRAG )
      nextsib = CONTENT_IS_PCDATA ;
    else
    {
      if ( FindElement( xml->GetNodeName() ) == 0 )
        nextsib = CurrentDTDNode ;
    }
    xml->FindNode( (const long)nIndex ) ;
  }
  else
    nsIndex = -1 ;

  if ( xml->MoveToPrevSibling() == 0 )
  {
    psIndex = xml->GetNodeIndex() ;
    if ( xml->NodeType() == NODE_FRAG )
      prevsib = CONTENT_IS_PCDATA ;
    else
    {
      if ( FindElement( xml->GetNodeName() ) == 0 )
        prevsib = CurrentDTDNode ;
    }
    xml->FindNode( (const long)nIndex ) ;
  }
  else
    psIndex = -1 ;

  if ( nodeptr != NULL ) CurrentDTDNode = nodeptr ;

  if ( parent == NULL && nextsib == NULL && prevsib == NULL )  /* Top of node tree */
  {  
    xml->FindNode( (const long)nIndex ) ;
    return(0) ;
  }

  if ( (ret = ValidateParentSibling( nodeptr, parent, nextsib, prevsib )) != 0 )
    return(ret) ;

  xml->FindNode( (const long)nIndex ) ;
  return(0) ;

}


int XML::DTDCheckDocNode( XML *xml )
{

  struct _BaseDTD *nodeptr, *parent, *nextsib, *prevsib ;
  long nIndex, pIndex, nsIndex, psIndex ;
  int ret ;


  if ( xml->DocNodeType() == NODE_FRAG )
  {
    nodeptr = NULL ;
  }
  else
  {
    if ( FindElement( xml->DocGetNodeName() ) != 0 ) 
      return(E_XML_ELEMENT) ;
    nodeptr = CurrentDTDNode ;
    if ( (ret = DocValidateNodeAttributes( nodeptr, xml )) != 0 ) return(ret) ;
    nodeptr = CurrentDTDNode ;
  }

  nIndex = xml->DocGetNodeIndex() ;
  parent = nextsib = prevsib = NULL ;

  if ( xml->DocNodeHasParent() == 1 )
  {
    xml->DocMoveToParent() ;
    pIndex = xml->DocGetNodeIndex() ;
    if ( FindElement( xml->DocGetNodeName() ) == 0 )
      parent = CurrentDTDNode ;
    xml->DocFindNode( (const long)nIndex ) ;
  }
  else
    pIndex = -1 ;

  if ( xml->DocMoveToNextSibling() == 0 )
  {
    nsIndex = xml->DocGetNodeIndex() ;
    if ( xml->DocNodeType() == NODE_FRAG )
      nextsib = CONTENT_IS_PCDATA ;
    else
    {
      if ( FindElement( xml->DocGetNodeName() ) == 0 )
        nextsib = CurrentDTDNode ;
    }
    xml->DocFindNode( (const long)nIndex ) ;
  }
  else
    nsIndex = -1 ;

  if ( xml->DocMoveToPrevSibling() == 0 )
  {
    psIndex = xml->DocGetNodeIndex() ;
    if ( xml->DocNodeType() == NODE_FRAG )
      prevsib = CONTENT_IS_PCDATA ;
    else
    {
      if ( FindElement( xml->DocGetNodeName() ) == 0 )
        prevsib = CurrentDTDNode ;
    }
    xml->DocFindNode( (const long)nIndex ) ;
  }
  else
    psIndex = -1 ;

  if ( nodeptr != NULL ) CurrentDTDNode = nodeptr ;

  if ( parent == NULL && nextsib == NULL && prevsib == NULL )  /* Top of node tree */
  {  
    xml->DocFindNode( (const long)nIndex ) ;
    return(0) ;
  }

  if ( (ret = ValidateParentSibling( nodeptr, parent, nextsib, prevsib )) != 0 )
    return(ret) ;

  xml->DocFindNode( (const long)nIndex ) ;
  return(0) ;

}


int XML::ValidateParentSibling( struct _BaseDTD *nodeptr, struct _BaseDTD *parent, struct _BaseDTD *nextsib, struct _BaseDTD *prevsib )
{

  int out, occ, frag ;
  struct _BaseDTD *tmpnode ;


//  frag = ( nodeptr == NULL ) ? 1 : 0 ;

  if ( nodeptr == NULL )
    frag = 1 ;
  else
    frag = 0 ;

  if ( parent != NULL )
  {
    if ( parent->ContentType != ELEMENT_CONTENT_MIXED && parent->ContentType != ELEMENT_CONTENT_KIDS )
      return(E_XML_DTDELEMENT) ;
    nodeptr = parent->FirstChild ;
    for( out = 0 ; ; )
    {
      switch( nodeptr->Type )
      {
        case NODE_CONTENT_TYPE:
          if ( nodeptr->ContType != CONTENT_IS_PCDATA )
          {
            if ( strcmp( (const char *)CurrentDTDNode->Name, (const char *)nodeptr->ContType->Name ) == 0 )
            {
              occ = nodeptr->Occurence ;
              out = 1 ;
              if ( nodeptr->ContentModel == CONTENT_MODEL_SEQ && nodeptr->NextNode != NULL )
              {
                if (  nextsib == NULL && CheckRequiredElements( nodeptr->NextNode ) == 1 )
                  return(E_XML_DTDBADSIB) ;
                else
                  if ( nextsib == CONTENT_IS_PCDATA ) out = 1 ;
              }
            }
            else
            {
              if ( nodeptr->ContentModel == CONTENT_MODEL_SEQ )
              {
                if ( prevsib == NULL && ( nodeptr->Occurence == ELEMENT_ONLY_ONCE || nodeptr->Occurence == ELEMENT_ONE_OR_MORE )  )
                  return(E_XML_DTDBADSIB) ;
                else
                  if ( prevsib == CONTENT_IS_PCDATA ) out = 1 ;
              }
            } 
          }
          else
          {
            if ( frag ) out = 1 ;
          }
          break ;

        case NODE_CONTENT_MODEL:
          tmpnode = nodeptr ;
          if ( nodeptr->FirstChild != NULL )
          {
            nodeptr = nodeptr->FirstChild ;
            if ( nodeptr->ContentModel == CONTENT_MODEL_CHOICE )
            {
              for( out = 0 ; out != 1 ; )
              {
                switch( nodeptr->Type )
                {
                  case NODE_CONTENT_TYPE:
                    if ( strcmp( (const char *)CurrentDTDNode->Name, (const char *)nodeptr->ContType->Name ) == 0 )
                    {
//                      nodeptr = nodeptr->Parent ;
                      out = 1 ;
                    }
                    break ;

                  default:
                    out = out ;  /* break point later! */
                    break ;
                }
                if ( nodeptr->NextNode != NULL && out != 1 )
                  nodeptr = nodeptr->NextNode ;
                else
                  break ;
              }
              if ( out == 0 ) 
                nodeptr = tmpnode ;
            }
            else
            {
               out = out ;   /* break point later! */
            }
          }
          else
            return(E_DTD_INTERNAL) ;    

          break ;

        default:
          return(E_DTD_INTERNAL) ;
      }

      if ( nodeptr->NextNode == NULL || out == 1 )
        break ;
      else
        nodeptr = nodeptr->NextNode ;
    }
  }
  else
    return(E_DTD_INTERNAL) ;

  if ( out == 0 ) 
    return(E_XML_DTDELEMENT) ;

  tmpnode = nodeptr ;
  if ( prevsib != NULL )
  {
    if ( frag == 0 && prevsib != CONTENT_IS_PCDATA && strcmp( (const char *)prevsib->Name, (const char *)nodeptr->ContType->Name ) == 0 )
    {
      if ( occ == ELEMENT_ONLY_ONCE || occ == ELEMENT_ZERO_OR_ONE )
        return(E_XML_DTDBADSIB) ;    
    }
    else
    {
      if ( nodeptr->Parent->Type == NODE_CONTENT_MODEL )
        nodeptr = nodeptr->Parent ;
      if ( nodeptr->PrevNode != NULL )
      {

        if ( nodeptr->PrevNode->Type == NODE_CONTENT_MODEL )
        {
          if ( nodeptr->PrevNode->FirstChild != NULL )
          {
            nodeptr = nodeptr->PrevNode->FirstChild ;
            if ( nodeptr->ContentModel == CONTENT_MODEL_CHOICE )
            {
              for( out = 0 ; out != 1 ; )
              {
                switch( nodeptr->Type )
                {
                  case NODE_CONTENT_TYPE:
                    if ( strcmp( (const char *)prevsib->Name, (const char *)nodeptr->ContType->Name ) == 0 )
                      out = 1 ;
                    break ;

                  default:
                    out = out ;  /* break point later! */
                    break ;
                }
                if ( nodeptr->NextNode != NULL )
                  nodeptr = nodeptr->NextNode ;
                else
                  break ;
              }
              if ( out == 0 )
                return(E_XML_DTDBADSIB) ;
            }
            else
            {
               out = out ;   /* break point later! */
            }
          }
          else
            return(E_DTD_INTERNAL) ;    
        }
        else
        {
          if ( nodeptr->PrevNode->Occurence == ELEMENT_ONLY_ONCE || nodeptr->PrevNode->Occurence == ELEMENT_ONE_OR_MORE)
            if ( prevsib != CONTENT_IS_PCDATA )
            {
              if ( strcmp( (const char *)prevsib->Name, (const char *)nodeptr->PrevNode->ContType->Name ) != 0 )
                return(E_XML_DTDBADSIB) ;
            }
            else
              if ( nodeptr->PrevNode->ContType != CONTENT_IS_PCDATA )
                return(E_XML_DTDBADSIB) ;
        }


/*
        if ( strcmp( (const char *)prevsib->Name, (const char *)nodeptr->PrevNode->ContType->Name ) != 0 )
          return(E_XML_DTDBADSIB) ;
*/
      }
      else
        if ( frag == 0 ) 
          return(E_XML_DTDBADSIB) ;
    }
  }

  nodeptr = tmpnode ;
  if ( nextsib != NULL )
  {
    if ( frag == 0 && nextsib != CONTENT_IS_PCDATA && strcmp( (const char *)nextsib->Name, (const char *)nodeptr->ContType->Name ) == 0 )
    {
      if ( occ == ELEMENT_ONLY_ONCE || occ == ELEMENT_ZERO_OR_ONE )
        return(E_XML_DTDBADSIB) ;    
    }
    else
    {
      if ( nodeptr->Parent->Type == NODE_CONTENT_MODEL )
        nodeptr = nodeptr->Parent ;
      if ( nodeptr->NextNode != NULL )
      {
        if ( nodeptr->NextNode->Type == NODE_CONTENT_MODEL )
        {
          if ( nodeptr->NextNode->FirstChild != NULL )
          {
            nodeptr = nodeptr->NextNode->FirstChild ;
            if ( nodeptr->ContentModel == CONTENT_MODEL_CHOICE )
            {
              for( out = 0 ; out != 1 ; )
              {
                switch( nodeptr->Type )
                {
                  case NODE_CONTENT_TYPE:
                    if ( strcmp( (const char *)nextsib->Name, (const char *)nodeptr->ContType->Name ) == 0 )
                      out = 1 ;
                    break ;

                  default:
                    out = out ;  /* break point later! */
                    break ;
                }
                if ( nodeptr->NextNode != NULL )
                  nodeptr = nodeptr->NextNode ;
                else
                  break ;
              }
              if ( out == 0 ) 
                return(E_XML_DTDBADSIB) ;
            }
            else
            {
               out = out ;   /* break point later! */
            }
          }
          else
            return(E_DTD_INTERNAL) ;    
        }
        else
        {
          if ( nodeptr->NextNode->Occurence == ELEMENT_ONLY_ONCE || nodeptr->NextNode->Occurence == ELEMENT_ONE_OR_MORE )
            if ( nextsib != CONTENT_IS_PCDATA && nodeptr->NextNode->ContType != CONTENT_IS_PCDATA )
            {
              if ( strcmp( (const char *)nextsib->Name, (const char *)nodeptr->NextNode->ContType->Name ) != 0 )
                return(E_XML_DTDBADSIB) ;
            }
            else
              if ( nodeptr->NextNode->ContType != CONTENT_IS_PCDATA && nextsib == CONTENT_IS_PCDATA)
                return(E_XML_DTDBADSIB) ;

        }
      }
      else
        if ( frag == 0 ) 
          return(E_XML_DTDBADSIB) ;
    }
  }

  return(0) ;

}


int XML::CheckRequiredElements( struct _BaseDTD *nodeptr )
{

  if ( nodeptr->ContType == CONTENT_IS_PCDATA ) return(0) ;
  for( ; ; )
  {
    if ( nodeptr->Occurence == ELEMENT_ONLY_ONCE || nodeptr->Occurence == ELEMENT_ONE_OR_MORE )
      return(1) ;
    if ( nodeptr->NextNode != NULL )
      nodeptr = nodeptr->NextNode ;
    else
      break ;
  }

  return(0) ;

}


/* Only here for ASP/VB/etc. programs that can't use pointers to class instances properly */
long XML::GetObjectAddress( void ) { return( (long)this ) ; }


char *XML::GetDTDElementName( void ){ return( CurrentDTDNode->Name ) ; }


int XML::ElementHasAttr( void ){return( (CurrentDTDNode->FirstAttr != NULL) ? 1 : 0 ) ; }


int XML::DocInitialise( const char *RootElement, XML *dtdptr )
{

  if ( dtdptr != NULL )
  {
    if ( ! SYS_DTDREADY(dtdptr->GetSystemState() ) ) return(E_DTD_READY) ;
    if ( dtdptr->FindElement( RootElement ) != 0 ) return(E_DOC_ELEMENT) ;
  }
  if ( Doc.Name != NULL ) return(E_DOC_INIT) ;
  Doc.Type = NODE_BRANCH ;
  if ( dtdptr != NULL )
  {
    DocGotDTD = 1 ;
    Doc.Name = dtdptr->GetDTDElementName() ;
  }
  else
    Doc.Name = strdup(RootElement) ;
  CurrentDocNode = &Doc ; 
  Doc_XMLVersion = strdup( XML_VERSION_STRING ) ;
  SystemState |= SYS_STATE_DOCINIT ;
  return(0) ;

}


int XML::DocAddSibling( const char *ElementName, XML *dtdptr )
{

  struct _Node *tmpnode, *nsnode ;


  if ( Doc.Name == NULL ) return(E_DOC_NOT_INIT) ;
  if ( CurrentDocNode == &Doc ) return(E_DOC_SIBLING) ;
  if ( dtdptr != NULL )
    if ( dtdptr->FindElement( ElementName ) != 0 ) return(E_DOC_ELEMENT) ;

  tmpnode = new _Node ;
  ZapDocNode(tmpnode) ;
  if ( dtdptr != NULL )
  {
    if ( dtdptr->GetContentType() == ELEMENT_CONTENT_EMPTY )
      tmpnode->Type = NODE_EMPTY ;
    else
      tmpnode->Type = NODE_LEAF ;
    tmpnode->Name = dtdptr->GetDTDElementName() ;
  }
  else
  {
    tmpnode->Type = NODE_LEAF ;
    tmpnode->Name = strdup(ElementName) ;
  }

  tmpnode->Parent = CurrentDocNode->Parent ;
  
  if ( (nsnode = CurrentDocNode->NextSibbling ) != NULL )
  {
    CurrentDocNode->NextSibbling = tmpnode ;
    tmpnode->PreviousSibbling = CurrentDocNode ;
    tmpnode->NextSibbling = nsnode ;
    nsnode->PreviousSibbling = tmpnode ;
  }
  else
  {
    CurrentDocNode->NextSibbling = tmpnode ;
    tmpnode->PreviousSibbling = CurrentDocNode ;
  }

  CurrentDocNode = tmpnode ;
  return(0) ;

}


int XML::DocAddFirstChild( const char *ElementName, XML *dtdptr )
{

  struct _Node *tmpnode, *childnode = NULL ;


  if ( Doc.Name == NULL ) return(E_DOC_NOT_INIT) ;
  if ( CurrentDocNode->Type == NODE_FRAG ) return(E_DOC_FIRSTCHILD) ;
  if ( dtdptr != NULL )
    if ( dtdptr->FindElement( ElementName ) != 0 ) return(E_DOC_ELEMENT) ;

  if ( CurrentDocNode->FirstChild != NULL ) childnode = CurrentDocNode->FirstChild ;

  tmpnode = new struct _Node ;
  ZapDocNode( tmpnode ) ;
  CurrentDocNode->FirstChild = tmpnode ;
  tmpnode->Parent = CurrentDocNode ;
  if ( dtdptr != NULL )
  {
    if ( dtdptr->GetContentType() == ELEMENT_CONTENT_EMPTY )
      tmpnode->Type = NODE_EMPTY ;
    else
      tmpnode->Type = NODE_LEAF ;
    tmpnode->Name = dtdptr->GetDTDElementName() ;
  }
  else
  {
    tmpnode->Type = NODE_LEAF ;
    tmpnode->Name = strdup(ElementName) ;
  }
  CurrentDocNode->Type = NODE_BRANCH ;
  if ( childnode != NULL )
  {
    childnode->PreviousSibbling = tmpnode ;
	  tmpnode->NextSibbling = childnode ;
  }
  CurrentDocNode = tmpnode ;
  return(0) ;

}


int XML::DOCAddCDATA( const char *cdata, int mode )
{

  struct _Node *tmpnode, *nsnode, *childnode = NULL ;


  if ( Doc.Name == NULL ) return(E_DOC_NOT_INIT) ;
  if ( CurrentDocNode == &Doc && mode == ADDCDATA_SIBLING ) return(E_DOC_SIBLING) ;
  if ( CurrentDocNode->Type == NODE_CDATA && mode == ADDCDATA_CHILD ) return(E_DOC_ADDCDATA) ;

  tmpnode = new _Node ;
  ZapDocNode(tmpnode) ;

  tmpnode->Type = NODE_CDATA ;
  tmpnode->Value = strdup(cdata) ;

  switch(mode)
  {
    case ADDCDATA_SIBLING:
      tmpnode->Parent = CurrentDocNode->Parent ;
      if ( (nsnode = CurrentDocNode->NextSibbling ) != NULL )
      {
        CurrentDocNode->NextSibbling = tmpnode ;
        tmpnode->PreviousSibbling = CurrentDocNode ;
        tmpnode->NextSibbling = nsnode ;
        nsnode->PreviousSibbling = tmpnode ;
      }
      else
      {
        CurrentDocNode->NextSibbling = tmpnode ;
        tmpnode->PreviousSibbling = CurrentDocNode ;
      }
      break ;

    case ADDCDATA_CHILD:
      if ( CurrentDocNode->FirstChild != NULL ) childnode = CurrentDocNode->FirstChild ;
      CurrentDocNode->FirstChild = tmpnode ;
      tmpnode->Parent = CurrentDocNode ;
      CurrentDocNode->Type = NODE_BRANCH ;
      if ( childnode != NULL )
      {
        childnode->PreviousSibbling = tmpnode ;
	      tmpnode->NextSibbling = childnode ;
      }
      break ;
    
    default:
      return(E_DOC_ADDCDATA) ;
      break ;
  }
  CurrentDocNode = tmpnode ;
  return(0) ;

}


int XML::DocAddFrag( const char *frag )
{

  struct _Node *tmpnode ;


  if ( Doc.Name == NULL ) return(E_DOC_NOT_INIT) ;
  if ( CurrentDocNode->Type == NODE_FRAG || ( CurrentDocNode->FirstChild != NULL && CurrentDocNode->FirstChild->Type == NODE_FRAG ) )
    return(E_DOC_ADDFRAG) ;

  tmpnode = new struct _Node ;
  ZapDocNode( tmpnode ) ;
  tmpnode->Parent = CurrentDocNode ;
  tmpnode->Type = NODE_FRAG ;
  CurrentDocNode->Type = NODE_BRANCH ;
  tmpnode->Name = NULL ;
  tmpnode->Value = strdup( frag ) ;
  if ( CurrentDocNode->FirstChild != NULL )
  {
    CurrentDocNode->FirstChild->PreviousSibbling = tmpnode ;
    tmpnode->NextSibbling = CurrentDocNode->FirstChild ;
  }
  CurrentDocNode->FirstChild = tmpnode ;
  return(0) ;

}


int XML::DocSetNodeValue( const char *nValue )
{
  if ( nValue == NULL ) return(E_DOC_SETVALUE) ;
  if ( CurrentDocNode->Value != NULL ) free(CurrentDocNode->Value) ;
//  CurrentDocNode->Value = EscapeSpecialChars( nValue ) ;
  CurrentDocNode->Value = strdup(nValue) ;
  return(0) ;

}


int XML::DocSetAttribute( const char *attrname, const char *attrval, XML *dtdptr)
{

  struct _Node *tmpnode ;
  int foundit ;


  if ( CurrentDocNode->Type == NODE_FRAG ) return(E_DOC_DTDATTR) ;
  if ( dtdptr != NULL )
  {
    if ( dtdptr->FindElement( (const char *)CurrentDocNode->Name ) != 0 ) return(E_DOC_INTERNAL) ;
    if ( dtdptr->ElementHasAttr() == 0 ) return(E_DOC_DTDATTR) ;
  }
  if ( CurrentDocNode->Attributes == NULL )
    return(DocAddAttrNode( attrname, attrval, (struct _Node *)NULL )) ;
  else
  {
    tmpnode = CurrentDocNode->Attributes ;
    for ( foundit = 0 ; ; )
    {
      if ( strcmp( attrname, (const char *)tmpnode->Name ) == 0 )
      {
        foundit = 1 ;
        free( tmpnode->Value );
        tmpnode->Value = strdup( attrval ) ;
        break ;
      }
      if ( tmpnode->NextSibbling != NULL )
        tmpnode = tmpnode->NextSibbling ;
      else
        break ;
    }
    if ( foundit == 0 )
      DocAddAttrNode( attrname, attrval, tmpnode ) ;
  }

  return(0) ;

}


int XML::DocAddAttrNode( const char *attrname, const char *attrval, struct _Node *lastsib )
{

  struct _Node *tmpnode ;


  tmpnode = new _Node ;
  ZapDocNode( tmpnode ) ;
  tmpnode->Parent = CurrentDocNode ;
  tmpnode->Name = strdup(attrname) ;
  tmpnode->Value = strdup(attrval) ;
  tmpnode->Type = NODE_ATTRVAL ;
  if ( lastsib != NULL )
  {
    lastsib->NextSibbling = tmpnode ;
    tmpnode->PreviousSibbling = lastsib ;
  }
  else
    CurrentDocNode->Attributes = tmpnode ;

  return(0) ;

}


int XML::DocNodeHasChild( void ){return( (CurrentDocNode->FirstChild != NULL) ? 1 : 0 ) ; }


int XML::DocNodeHasSibling( void ){return( (CurrentDocNode->NextSibbling != NULL) ? 1 : 0 ) ; }


int XML::DocNodeHasParent( void ){return( (CurrentDocNode->Parent != NULL) ? 1 : 0 ) ; }


int XML::DocNodeHasAttr( void ){return( (CurrentDocNode->Attributes != NULL) ? 1 : 0 ) ; }


long XML::DocGetNodeIndex( void ){ return( CurrentDocNode->Index ) ; }


const int XML::DocNodeType( void )
{

  return( (const int )CurrentDocNode->Type ) ;

}


const char *XML::DocGetNodeAttrValue( const char *at )
{

  struct _Node *tmpnode ;

  if ( CurrentDocNode->Attributes != NULL )
  {
     tmpnode = CurrentDocNode->Attributes ;
     for ( ; ; )
     {
       if ( stricmp( at, (const char *)tmpnode->Name ) == 0 )
         return( (const char *)tmpnode->Value ) ;
       if ( tmpnode->NextSibbling != NULL )
         tmpnode = tmpnode->NextSibbling ;
       else
         break ;
     }
  }

  return( (const char *)NULL ) ;

}


const char *XML::DocGetParentName( void )
{

  if ( CurrentDocNode->Parent != NULL )
    return((const char *)CurrentDocNode->Parent->Name ) ;
  else
    return((const char *)"") ;

}


long XML::DocGetParentIndex( void )
{

  if ( CurrentDocNode->Parent != NULL )
    return( CurrentDocNode->Parent->Index ) ;
  else
    return(-1) ;

}


const char *XML::DocGetNodeName()
{

  return((const char *)CurrentDocNode->Name ) ;

}


int XML::DocMoveToFirstChild( void )
{

  if ( CurrentDocNode->FirstChild != NULL ) 
  {
    CurrentDocNode = CurrentDocNode->FirstChild ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::DocMoveToNextSibling( void )
{

  if ( CurrentDocNode->NextSibbling != NULL ) 
  {
    CurrentDocNode = CurrentDocNode->NextSibbling ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::DocMoveToParent( void )
{

  if ( CurrentDocNode->Parent != NULL ) 
  {
    CurrentDocNode = CurrentDocNode->Parent ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::DocMoveToPrevSibling( void )
{

  if ( CurrentDocNode->PreviousSibbling != NULL ) 
  {
    CurrentDocNode = CurrentDocNode->PreviousSibbling ;
    return(0) ;
  }
  else
    return(1) ;

}


int XML::DocFindNode( const long NodeIndex )
{

  long i ;


  if ( NodeIndex <= DocIndex )
  {
    CurrentDocNode = &Doc ;
    for ( i = 0 ; i < NodeIndex ; i++ )
      CurrentDocNode = CurrentDocNode->NextIndex ;
    return(0) ;
  }
  return(1) ;

}


int XML::DocFindNode( const char *NodeName, const long Occurence )
{

  int i ;


  CurrentDocNode = &Doc ;
  i = 0 ;
  do
  {
    
    if ( stricmp( NodeName, CurrentDocNode->Name ) == 0 )
    {
      i++ ;
      if ( i == Occurence ) return(0) ;
    }
    if ( CurrentDocNode->FirstChild != NULL )
      CurrentDocNode = CurrentDocNode->FirstChild ;
    else
      if ( CurrentDocNode->NextSibbling != NULL ) 
        CurrentDocNode = CurrentDocNode->NextSibbling ;
      else
      {
        CurrentDocNode = CurrentDocNode->Parent ;
        if ( CurrentDocNode->NextSibbling != NULL )
          CurrentDocNode = CurrentDocNode->NextSibbling ;
        else
        {
          while( CurrentDocNode->NextSibbling == NULL && CurrentDocNode->Parent != NULL ) 
            CurrentDocNode = CurrentDocNode->Parent ;
          if ( CurrentDocNode->Parent != NULL ) CurrentDocNode = CurrentDocNode->NextSibbling ;
        }
      }
  } while ( CurrentDocNode->Parent != NULL ) ;

  return( 1 ) ;

}


int XML::DocFindNode( const int NodeType, const long Occurence )
{

  int i ;


  CurrentDocNode = &Doc ;
  i = 0 ;
  do
  {
    
    if ( NodeType == CurrentDocNode->Type )
    {
      i++ ;
      if ( i == Occurence ) return(0) ;
    }
    if ( CurrentDocNode->FirstChild != NULL )
      CurrentDocNode = CurrentDocNode->FirstChild ;
    else
      if ( CurrentDocNode->NextSibbling != NULL ) 
        CurrentDocNode = CurrentDocNode->NextSibbling ;
      else
      {
        CurrentDocNode = CurrentDocNode->Parent ;
        if ( CurrentDocNode->NextSibbling != NULL )
          CurrentDocNode = CurrentDocNode->NextSibbling ;
        else
        {
          while( CurrentDocNode->NextSibbling == NULL && CurrentDocNode->Parent != NULL ) 
            CurrentDocNode = CurrentDocNode->Parent ;
          if ( CurrentDocNode->Parent != NULL ) CurrentDocNode = CurrentDocNode->NextSibbling ;
        }
      }
  } while ( CurrentDocNode->Parent != NULL ) ;

  return( 1 ) ;

}


int XML::DocFindNode( const char *NodeName, const char *AttrName, const char *AttrValue )
{


  CurrentDocNode = &Doc ;
  do
  {
    
    if ( strcmp( (const char *)CurrentDocNode->Name, NodeName) == 0 && 
         DocNodeHasAttribute(AttrName) &&
         strcmp(DocGetNodeAttrValue(AttrName), AttrValue) == 0
       )
    {
      return(0) ;
    }
    if ( CurrentDocNode->FirstChild != NULL )
      CurrentDocNode = CurrentDocNode->FirstChild ;
    else
      if ( CurrentDocNode->NextSibbling != NULL ) 
        CurrentDocNode = CurrentDocNode->NextSibbling ;
      else
      {
        CurrentDocNode = CurrentDocNode->Parent ;
        if ( CurrentDocNode->NextSibbling != NULL )
          CurrentDocNode = CurrentDocNode->NextSibbling ;
        else
        {
          while( CurrentDocNode->NextSibbling == NULL && CurrentDocNode->Parent != NULL ) 
            CurrentDocNode = CurrentDocNode->Parent ;
          if ( CurrentDocNode->Parent != NULL ) CurrentDocNode = CurrentDocNode->NextSibbling ;
        }
      }
  } while ( CurrentDocNode->Parent != NULL ) ;

  return( 1 ) ;

}


void XML::DocSetDocType( const char *dt )
{

  if ( Doc_DocType != NULL )
  {
    free(Doc_DocType) ;
    Doc_DocType = NULL ;
  }
  if ( dt != NULL )
    Doc_DocType = strdup( dt ) ;

}


const char *XML::DocGetDocType( void )
{

  if ( Doc_DocType == NULL )
    return("") ;
  else
    return((const char *)Doc_DocType) ;

}


int XML::DocMake( void )
{

  unsigned int memsize ;
//  struct _Node *tmpnode ;
  int indent = 0 ;
  int finished = 0 ;
  char *p ;


  if ( ! SYS_DOCINIT ) return(E_DOC_READY) ;
  if ( XMLDoc != NULL ) free(XMLDoc) ;
  XMLDoc = (char *)malloc( MALLOC_INITIAL_SIZE ) ;
  if ( XMLDoc == NULL ) return(E_DOC_MALLOC) ;
  memsize = MALLOC_INITIAL_SIZE ;
  p = XMLDoc ;
  sprintf( XMLDoc, "%s\n", XML_VERSION_STRING ) ;
  p = p + strlen(XML_VERSION_STRING) + 1 ;
  if ( Doc_DocType != NULL )
  {
    if ( strlen((const char *)Doc_DocType) > 0 )
    {
      sprintf( p, "<!DOCTYPE %s>\n", Doc_DocType ) ;
      p = p + strlen((const char *)p) ;
    }
  }
  CurrentDocNode = &Doc ;
  do
  {
    if ( CurrentDocNode->Deleted == 0 )
    {
      p += DocMakeLine( p, indent ) ;
    }
    if ( CurrentDocNode->FirstChild != NULL && CurrentDocNode->Deleted == 0 )
    {
      *p++ = 13 ; *p++ = 10 ; *p = 0 ;    
      CurrentDocNode = CurrentDocNode->FirstChild ;
      indent += 2 ;
    }
    else
    {
      if ( CurrentDocNode->Type != NODE_EMPTY && CurrentDocNode->Deleted == 0 ) p += DocMakeEndTag( p, 0) ;
      *p++ = 13 ; *p++ = 10 ; *p = 0 ;    
      if ( CurrentDocNode->NextSibbling != NULL )
        CurrentDocNode = CurrentDocNode->NextSibbling ;
      else
      {
        CurrentDocNode = CurrentDocNode->Parent ;
        indent -= 2 ;
        if ( CurrentDocNode->Type != NODE_EMPTY && CurrentDocNode->Deleted == 0) p += DocMakeEndTag( p, indent) ;
        *p++ = 13 ; *p++ = 10 ; *p = 0 ;    
        if ( CurrentDocNode->NextSibbling != NULL )
          CurrentDocNode = CurrentDocNode->NextSibbling ;
        else
        {
          while ( CurrentDocNode->NextSibbling == NULL && CurrentDocNode->Parent != NULL )
          {
            CurrentDocNode = CurrentDocNode->Parent ;
            indent -= 2 ;
            if ( CurrentDocNode->Type != NODE_EMPTY && CurrentDocNode->Deleted == 0) p += DocMakeEndTag( p, indent) ;
            *p++ = 13 ; *p++ = 10 ; *p = 0 ;    
          }
          if ( CurrentDocNode->Parent != NULL && CurrentDocNode->NextSibbling != NULL )
            CurrentDocNode = CurrentDocNode->NextSibbling ;
        }
      }
    }
  }
  while ( CurrentDocNode->Parent != NULL ) ;

  return(0) ;

}


unsigned int XML::DocMakeLine( char *outline, int indent )
{

  const char *p ;
  char *indentstr ;
  struct _Node *tmpnode ;
  int noval = 0 ;


 // if ( CurrentDocNode->Type == NODE_FRAG ) 
 //   indent -= 2 ;
  indentstr = (char *)malloc( indent + 1 ) ;
  indentstr[indent] = 0 ;
  while ( indent-- ) indentstr[indent] = 32 ;
  switch( CurrentDocNode->Type )
  {
    case NODE_FRAG:
      sprintf( outline, "%s", indentstr ) ;
      break ;

    case NODE_CDATA:
      sprintf( outline, "%s%s", indentstr, CDATA_START) ;
      break ;

    default:
      sprintf( outline, "%s<%s", indentstr, CurrentDocNode->Name ) ;
      break ;
  }

  free(indentstr) ;
/*
  p = CurrentDocNode->AttrList ;
  if ( p != NULL )
  {
    if ( strlen( p ) > 0 )
    {
      strcat( outline, p ) ;
    }
  }
*/
  if ( CurrentDocNode->Attributes != NULL )
  {
    tmpnode = CurrentDocNode->Attributes ;
    for( ; ; )
    {
      p = outline + strlen((const char *)outline) ;
      sprintf((char *)p, " %s=\"%s\"", tmpnode->Name, tmpnode->Value ) ;      
      if ( tmpnode->NextSibbling != NULL )
        tmpnode = tmpnode->NextSibbling ;
      else
        break ;
    }
  }
  if ( CurrentDocNode->Type != NODE_FRAG &&
       CurrentDocNode->Type != NODE_CDATA ) 
  {
    if ( NodeHasNoValue((const char *)CurrentDocNode->Value) == 1 && CurrentDocNode->FirstChild == NULL )
    {
      if ( CurrentDocNode->Type == NODE_EMPTY )
        strcat( outline, ">" ) ;
      else
        strcat( outline, "/>" ) ;
    }
    else
      strcat( outline, ">" ) ;
  }
  if ( CurrentDocNode->Value != NULL )
  {
    p = EscapeSpecialChars((const char *)CurrentDocNode->Value) ;
    if ( p != NULL )
      strcat( outline, p ) ;
  }
  return(strlen((const char *)outline)) ;

}


int XML::NodeHasNoValue( const char *NodeVal )
{

  if ( NodeVal == NULL )
    return(1) ;
  else
  {
    if ( strlen(NodeVal) == 0 )
      return(1);
  }

  return(0) ;

}


unsigned int XML::DocMakeEndTag( char *outline, int indent)
{

  char *indentstr ;


  if ( CurrentDocNode->Type == NODE_FRAG ) return(0) ;
  indentstr = (char *)malloc( indent + 1 ) ;
  indentstr[indent] = 0 ;
  while ( indent-- ) indentstr[indent] = 32 ;

  sprintf( outline, (const char *)indentstr ) ;
  free(indentstr) ;
  if ( CurrentDocNode->Type == NODE_CDATA )
  {
    strcat(outline,CDATA_END) ;
  }
  else
  {
    if ( NodeHasNoValue((const char *)CurrentDocNode->Value) == 0 || CurrentDocNode->FirstChild != NULL )
    {
      strcat( outline, "</") ;
      strcat( outline, (const char *)CurrentDocNode->Name ) ;
      strcat( outline, ">" ) ;
    }
  }

  return(strlen((const char *)outline)) ;
  
}


const char *XML::GetXMLDoc( void )
{

  if ( XMLDoc == NULL )
    return((const char *)"") ;
  else
    return((const char *)XMLDoc ) ;

}


char *XML::EscapeSpecialChars( const char *str )
{

  char *p, *q ;
  char escstr[ESC_CHAR_LEN] ;
  int i ;

  
  p = strdup(str) ;
  for ( i = 0 ; i < NUMESCAPES ; i++ )
  { 
    escstr[0] = esc[i].eChar ;
    escstr[1] = 0 ;
    q = Replace( p, escstr, (char *)esc[i].eSeq ) ;
    free(p) ;
    p = q ;
  }

  return(p) ;

}


void XML::DocMoveToTop( void ){ CurrentDocNode = &Doc ; }


const char *XML::DocNodeValue( void )
{

  return( (const char *)CurrentDocNode->Value ) ;

}


int XML::CopyXMLToDoc( void )
{

  int ret ;


  if ( ! SYS_XMLREADY(SystemState) ) return(E_DOC_COPYNOXML) ;
  DocWipe() ;
  Doc_XMLVersion = strdup( XML_VERSION_STRING ) ;
  if ( DocType != NULL )
  {
    if ( strlen((const char *)DocType ) > 0 )
      Doc_DocType = strdup((const char *)DocType) ;
  }
  CurrentNode = &Top ;
  CurrentDocNode = &Doc ;
  ZapDocNode( &Doc ) ;

    if ( (ret = CopyNode(COPY_NODE_ROOT)) != 0 ) return(ret) ;
    do
    {
      if ( CurrentNode->FirstChild != NULL )
      {
        CurrentNode = CurrentNode->FirstChild ;
        if ( (ret = CopyNode(COPY_NODE_CHILD)) != 0 ) return(ret) ;
      }
      else
      {
        if ( CurrentNode->NextSibbling != NULL )
        {
          CurrentNode = CurrentNode->NextSibbling ;
          if ( (ret = CopyNode(COPY_NODE_SIBLING)) != 0 ) return(ret) ;
        }
        else
        {
          CurrentNode = CurrentNode->Parent ;
          CurrentDocNode = CurrentDocNode->Parent ;
          if ( CurrentNode->NextSibbling != NULL )
          {
            CurrentNode = CurrentNode->NextSibbling ;
            if ( (ret = CopyNode(COPY_NODE_SIBLING)) != 0 ) return(ret) ;
          }
          else
          {
            while ( CurrentNode->NextSibbling == NULL && CurrentNode->Parent != NULL )
            {
              CurrentNode = CurrentNode->Parent ;
              CurrentDocNode = CurrentDocNode->Parent ;
            }
            if ( CurrentNode->Parent != NULL && CurrentNode->NextSibbling != NULL )
            {
              CurrentNode = CurrentNode->NextSibbling ;
              if ( (ret = CopyNode(COPY_NODE_SIBLING)) != 0 ) return(ret) ;
            }
          }
        }
      }
    }
    while ( CurrentNode->Parent != NULL ) ;

    SystemState |= SYS_STATE_DOCINIT ;
    return(0) ;

}


int XML::CopyNode( int CopyMode )
{

  int ret = 0 ;
  struct _Node *tmpnode ;


  switch( CopyMode )
  {
    case COPY_NODE_ROOT:
      ret = CopyNodeCommon() ;
      break ;

    case COPY_NODE_CHILD:
      tmpnode = new _Node ;
      ZapDocNode( tmpnode ) ;
      CurrentDocNode->FirstChild = tmpnode ;
      tmpnode->Parent = CurrentDocNode ;
      CurrentDocNode = tmpnode ;
      ret = CopyNodeCommon() ;
      break ;

    case COPY_NODE_SIBLING:
      tmpnode = new _Node ;
      ZapDocNode( tmpnode ) ;
      CurrentDocNode->NextSibbling = tmpnode ;
      tmpnode->Parent = CurrentDocNode->Parent ;
      tmpnode->PreviousSibbling = CurrentDocNode ;
      CurrentDocNode = tmpnode ;
      ret = CopyNodeCommon() ;
      break ;
  }

  return(ret) ;

}


int XML::CopyNodeCommon( void )
{
  
  int ret = 0 ;


  CurrentDocNode->Type = CurrentNode->Type ;
  CurrentDocNode->Name = strdup((const char *)CurrentNode->Name) ;
  if ( CurrentNode->Value != NULL )
    CurrentDocNode->Value = strdup((const char *)CurrentNode->Value) ;
  if ( CurrentNode->Attributes != NULL )
    ret = CopyAttrNodes() ;

  return(ret) ;

}


int XML::CopyAttrNodes( void )
{

  struct _Node *tmpnode, *nodeptr, *lastattr ;


  nodeptr = CurrentNode->Attributes ;
  tmpnode = new _Node ;
  ZapDocNode( tmpnode ) ;
  CurrentDocNode->Attributes = tmpnode ;
  tmpnode->Parent = CurrentDocNode ;
  tmpnode->Type = NODE_ATTRVAL ;
  tmpnode->Name = strdup((const char *)nodeptr->Name) ;
  tmpnode->Value = strdup((const char *)nodeptr->Value) ;
  lastattr = tmpnode ;
  while( nodeptr->NextSibbling != NULL )
  {
    nodeptr = nodeptr->NextSibbling ;
    tmpnode = new _Node ;
    ZapDocNode( tmpnode ) ;
    tmpnode->Parent = CurrentDocNode ;
    tmpnode->PreviousSibbling = lastattr ;
    lastattr->NextSibbling = tmpnode ;
    tmpnode->Type = NODE_ATTRVAL ;
    tmpnode->Name = strdup((const char *)nodeptr->Name) ;
    tmpnode->Value = strdup((const char *)nodeptr->Value) ;
    lastattr = tmpnode ;
  }

  return(0) ;

}


int XML::DocDeleteNode( void )
{

  if ( CurrentDocNode == &Doc )
    return(E_DOC_DELETE) ;
  else
    CurrentDocNode->Deleted = 1 ;

  return(0) ;

}


void XML::DocUnDeleteNode( void )
{

  CurrentDocNode->Deleted = 0 ;

}


const int XML::NodeDeleted( void ) { return((const int)CurrentDocNode->Deleted) ; }
